#$Id: FileMan.pm,v 0.06 2000/12/15 16:00:00 george@xorgate.com Exp $
package Apache::FileMan;
use strict;

$Apache::FileMan::VERSION = '0.06';

use Apache::Constants qw(:common OPT_INDEXES DECLINE_CMD REDIRECT DIR_MAGIC_TYPE);
use DynaLoader ();
use Fcntl qw/:flock/;
use Apache::Util qw(ht_time size_string escape_html);
use Apache::ModuleConfig;
use Apache::Icon;
use Apache::Language;
use Apache::Request;
use Apache::Log;

#Configuration constants
use constant FANCY_INDEXING 	=> 1;
use constant ICONS_ARE_LINKS 	=> 2;
use constant SCAN_HTML_TITLES 	=> 4;
use constant SUPPRESS_LAST_MOD	=> 8;
use constant SUPPRESS_SIZE  	=> 16;
use constant SUPPRESS_DESC 		=> 32;
use constant SUPPRESS_PREAMBLE 	=> 64;
use constant SUPPRESS_COLSORT 	=> 128;
use constant THUMBNAILS 		=> 256;
use constant SHOW_PERMS         => 512;
use constant NO_OPTIONS			=> 1024;
use constant SKIP_INDEX			=> -1;
use constant HOLD_ERR			=> -2;
use constant ROOT_ERR			=> -3;

use vars qw(%GenericDirectives);
%GenericDirectives = 
(      
fancyindexing         =>  FANCY_INDEXING,
iconsarelinks         =>  ICONS_ARE_LINKS,
scanhtmltitles        =>  SCAN_HTML_TITLES,
suppresslastmodified  =>  SUPPRESS_LAST_MOD,
suppresssize          =>  SUPPRESS_SIZE,
suppressdescription   =>  SUPPRESS_DESC,
suppresshtmlperamble  =>  SUPPRESS_PREAMBLE,
suppresscolumnsorting =>  SUPPRESS_COLSORT,
thumbnails            =>  THUMBNAILS,
showpermissions       =>  SHOW_PERMS,
);

#Default values
use constant DEFAULT_ICON_WIDTH => 20;
use constant DEFAULT_ICON_HEIGHT=> 22;
use constant DEFAULT_NAME_WIDTH => 23;
use constant DEFAULT_ORDER		=> 'ND';
use constant DEFAULT_ROOT_DIR 	=> '.XFM';
use constant DEFAULT_HOLD_DIR 	=> 'hold';
use constant DEFAULT_TEXT_LEN 	=> 49;
use constant DEFAULT_HOLD_MENU => ['Upload','Unzip','Delete','MkDir','MkFile','Copy','Move','Edit','Rename'];
use constant DEFAULT_ROOT_MENU => ['Upload','Unzip','Delete','MkDir','MkFile','Copy','Move','Edit','Rename','Help'];
use constant DEFAULT_ADMN_MENU => ['SetGID','Revoke','Debug',];
use constant DEFAULT_POST_MAX 	=> 4194304;
use constant DEFAULT_EDIT_MAX 	=>  131072;
use constant DEFAULT_HELP_URL 	=> 'http://www.xorgate.com/help/FileMan';
use constant DEFAULT_DIR_MOD 	=> 0770;
use constant DEFAULT_FILE_MOD 	=> 0460;
use constant REVOKE_DIR			=> '/revoke';
use constant REVOKE_FILE		=> '/revoked';

#Global Options/Congiguration Directives
use vars qw($config);
             
use vars qw(%sortname);
%sortname =
( 	
'N'=>'Name' ,
'M'=>'LastModified',
'S'=>'Size',
'D'=>'Description',
);


# my global arguments
my ($debug,$dodump,$errmsg,$fmroot,$hold,$admin,$user,$gidname,$gid,$chgid,$umask);

# global users revoke cache
my $Users;

#Statistics variables
use vars qw($nDir $nRedir $nIndex $nThumb);
$nDir=0;
$nRedir=0;
$nIndex=0;
$nThumb=0;

if ($ENV{MOD_PERL}){
    no strict;
    @ISA=qw(DynaLoader);
    __PACKAGE__->bootstrap($Apache::FileMan::VERSION);
    if (Apache->module('Apache::Status')) {
	Apache::Status->menu_item('FileMan'=>'Apache::FileMan status',\&status);
    }
}

sub file_man {
    my($r,$args,$mode) = @_;
    my $name=$r->filename();
    my $holdname;
    my $docroot=$r->document_root;
    my $uri = $r->uri;
    my $cfg = Apache::ModuleConfig->get($r);
    my $subr;
    my $isroot;
    my $title;
    my $retval;
    my $msg='';
    my $rootdir=$cfg->{rootdir};
    my $holddir=$cfg->{holddir};
    my $lang = new Apache::Language($r) if $cfg->{language};
    print STDERR "file_man() $name\n" if $debug;
    if($mode=~m:h:o) {
	$holdname = $name;
	($args->{root_dir})=$uri=~m:(.*/)$holddir/: unless $args->{root_dir};
	$name=$docroot.$args->{root_dir};
	$name=~s:(.*)/$rootdir(/.*):$1$2:;	# strip out the root directory
    } elsif($mode=~m:x:o) {
	unless($args->{hold_dir}) {
	    $uri=~m:(.*/$rootdir/):;
	    $args->{hold_dir}="$1$holddir/";
	}
	$holdname=$docroot.$args->{hold_dir} if $hold;
    } 
    if($mode) {
	$msg=$lang->{FileMan} || 'FileMan';
    } else {
	$msg=$lang->{Header} || 'Directory index of';
    }
    if($mode) {
    	$args->{root_dir}=~m:(.*)${rootdir}/(.*):;
    	$title="$1$2";
    } else {
	$title=$uri;
    }
    $msg.=" $title";
    thumb_conf($r) if $cfg->{options} & THUMBNAILS;
    return OK unless header($r,$msg);
    if($holdname) {
	$r->filename($holdname);
	if(opendir HDH, $holdname) {
	    $msg=$lang->{HoldHeader} || 'HOLD index of';
	    $isroot=$holdname=~m:$rootdir/$holddir/$:;
	    print STDERR "file_man(hold) open hold_dir=$args->{hold_dir}\n" if $debug;
	    print qq~<H3><A NAME="hold">$msg $args->{hold_dir}</A></H3>\n~,
		qq~<FORM METHOD="POST" ACTION="$uri" ENCTYPE="MULTIPART/FORM-DATA">\n~;
	    if($cfg->{options} & FANCY_INDEXING) {
		hold_form($r,$args,$cfg,1,$cfg->{holdmenu}||DEFAULT_HOLD_MENU);
		fancy_page($r,$args,\*HDH,'hold',$isroot); 
	    } else {
		hold_form($r,$args,$cfg,0,$cfg->{holdmenu}||DEFAULT_HOLD_MENU);
		plain_page($r,$args,\*HDH,'hold',$isroot);
	    }
	    print "</FORM>\n";
	    closedir HDH;
	} else {
	    $r->warn(__PACKAGE__ . " Can't open directory $holdname");
	}
    }
    if($mode) {
	$msg=$lang->{RootHeader} || 'ROOT index of';
    } else {
	$msg=$lang->{Header} || 'Directory index of';
    }
    print qq~<H3><A NAME="root">$msg $title</A></H3>\n~;
    if($mode) {
	print qq~<FORM METHOD="POST" ACTION="$uri" ENCTYPE="MULTIPART/FORM-DATA">\n~;
	$isroot=$uri=~m:^$fmroot$rootdir/$:;
    } else {
	$isroot=$uri=~m:^$fmroot$:;
    }
    $r->filename("$name");
    unless (opendir DH, $name) {
	$r->log_reason( __PACKAGE__ . " Can't open directory $name");
	return FORBIDDEN;
    }
    $nDir++;
    if($cfg->{options} & FANCY_INDEXING) {
	root_form($r,$args,$cfg,1,$cfg->{rootmenu}||DEFAULT_ROOT_MENU) if $mode;
	$retval=fancy_page($r,$args,\*DH,$mode?'root':'',$isroot);
	print $ENV{'SERVER_SIGNATURE'};
    } else {
	root_form($r,$args,$cfg,0,$cfg->{rootmenu}||DEFAULT_ROOT_MENU)
	if $mode;
	$retval=plain_page($r,$args,\*DH,$mode?'root':'',$isroot);
    }
    print "</FORM>\n" if $mode;
    closedir DH;
    return $retval;
}

sub header {
    my ($r,$title)=@_;
    my $subr;
    my $htmlop;
    my $cfg = Apache::ModuleConfig->get($r);
    $r->no_cache(1) if $cfg->{nocache};
    $r->send_http_header("text/html");
    return 0 if $r->header_only;
    print qq~<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.0 Transitional//EN""http://www.w3.org/TR/REC-html40/loose.dtd">\n~;
    print "<HTML><HEAD><TITLE>$title</TITLE></HEAD>\n";
    $htmlop=$r->dir_config('IndexHtmlBody');
    if($htmlop){
	$config->{IndexHtmlBody}=$htmlop;	# Record for debug Dumper
	eval 'print "$htmlop"';
    } else {
	print "<BODY>";
    }
    $htmlop=$r->dir_config('IndexHtmlHead');
    if($htmlop) {
    $config->{IndexHtmlHead}=$htmlop;	# Record for debug Dumper
	$subr = $r->lookup_uri($htmlop);
	$subr->run;
    }
    if(@$gid) {
	print "<H3>";
	print qq~User="$user" ~ if $user;
	my $cnt=@$gid-1;
	print "FileMan Access=$gidname->[$cnt]";
	for($cnt--;$cnt>=0;$cnt--) {
	    print ",$gidname->[$cnt]";
	}
	print "</H3>\n";
    }
    place_doc($r,$cfg,'header') if $cfg->{options} & FANCY_INDEXING;
    1;
}

sub footer {
    my ($r)=@_;
    my $htmlop;
    my $cfg = Apache::ModuleConfig->get($r);
    place_doc($r,$cfg,'readme') if $cfg->{options} & FANCY_INDEXING;
    $htmlop=$r->dir_config('IndexHtmlFoot');
    if($htmlop) {
	$config->{IndexHtmlFood}=$htmlop;	# Record for debug Dumper
	my $subr = $r->lookup_uri($htmlop);
	$subr->run;
    }
    print "</BODY></HTML>\n";
}

sub nsubstr {
    my ($str,$substr,$offset) =@_;
    my ($cnt,$ndx);
    my $sublen=length $substr;
    for($cnt=0; ($ndx=index($str,$substr,$offset))>=0; $cnt++) {
	$offset=$ndx+$sublen;
    }
    $cnt;
}

sub dirbound {
    my ($dir, $root)=@_;
    my $level=nsubstr($root,'/');
    my $cnt=0;
    while($dir=~m:/:go) {
	$cnt++;
	if($dir=~m:\G\.\.(/|$):o) {
	    return 0 if --$cnt<$level;
	    $cnt-- if m:\G\.\./:o;
	}
    }
    1;
}

sub procform {
    my ($r,$args,$items) = @_;
    my $cfg = Apache::ModuleConfig->get($r);
    my $rootdir = $cfg->{rootdir};
    my $lang = new Apache::Language($r) if $cfg->{language};
    my $cmd;
    my $dst;
    my $msg='';
    my $dir;
    my $redir;
    my $file;
    my $child;
    my $formsrc;
    my $formdst;
    my $load;
    my $target;
    my $error;
    my $docroot=$r->document_root;
    my $count;
    my $proc=$args->{proc};
    my $retval=0;
    if($proc eq 'hold') {
    	$cmd = getcmd($cfg->{holdmenu},$args);
	if($args->{hold_src}=~m:^/:o) {
	    $formsrc="$args->{hold_src}";
	} else {
	    $formsrc="$args->{hold_dir}$args->{hold_src}";
	}
	if($args->{hold_dst}=~m:^/:o) {
	    $formdst="$args->{hold_dst}";
	} else {
	    $formdst="$args->{hold_dir}$args->{hold_dst}";
	}
	$redir	="$args->{hold_dir}";
	$dir	="$args->{hold_dir}";
	$file	="$args->{hold_file}";
	$child	="$args->{hold_child}";
	$load	="$args->{browse}";
	$error	=HOLD_ERR;
    } elsif($proc eq 'root') {
    	$cmd = getcmd($cfg->{rootmenu},$args);
	if($args->{root_src}=~m:^/:o) {
	    $formsrc="$args->{root_src}";
	} else {
	    ($formsrc="$args->{root_dir}$args->{root_src}")=~s:/$rootdir/:/:;
	}
	if($args->{root_dst}=~m:^/:o) {
	    $formdst="$args->{root_dst}";
	} else {
	    ($formdst="$args->{root_dir}$args->{root_dst}")=~s:/$rootdir/:/:;
	}
	$redir	="$args->{root_dir}";
	($dir=$redir)=~s:/$rootdir/:/:;
	$file	="$args->{root_file}";
	$child	="$args->{root_child}";
	$load	="$args->{browse}";
	$error	=ROOT_ERR;
    } else {
	$msg=$lang->{process} || 'UNKNOWN: process=';
	errmsg("$msg$proc");
	return 0;
    }
    $cmd||=getcmd($cfg->{admnmenu},$args);
    unless($cmd) {
	$msg=$lang->{command} || 'UNKNOWN: command';
	errmsg($msg);
	$r->log->error(__PACKAGE__." internal error: UNKNOWN command");
	return 0;
    }
    $dir    =~tr{ .a-zA-Z0-9~!@#$^&+i_\-/}{}cd; #strip unusual characters
    $file   =~tr{ .a-zA-Z0-9~!@#$^&+i_\-/}{}cd;
    $child  =~tr{ .a-zA-Z0-9~!@#$^&+i_\-/}{}cd;
    $formsrc=~tr{ .a-zA-Z0-9~!@#$^&+i_\-/}{}cd;
    $formdst=~tr{ .a-zA-Z0-9~!@#$^&+i_\-/}{}cd;
    print STDERR "procform($cmd) " if $debug;
    if($cmd eq 'Help') {
	$dst=$cfg->{help}||DEFAULT_HELP_URL;
	$dst.="?version=$Apache::FileMan::VERSION&postmax=$cfg->{postmax}";
	$dst.="&hold=1"   if $hold;
	$dst.="&perms=1"  if @$gid;
	$dst.="&admin=1"  if isadmin($admin);
	$msg="$user: $cmd: $dst";
	$r->log->notice(__PACKAGE__." $msg");
	$r->header_out(Location=>$dst);
	return REDIRECT;
    }
    if($cmd eq 'Debug') {
	$dodump = !$dodump if $debug;
	print STDERR "Debug=$dodump\n" if $debug;
	return 0;
    }
# Command only requires a $load variable
    if($cmd eq 'Upload') {
	if(!$load) {
	    errmsg('No upload file. (Hint: click "Browse...")');
	    return $error;
	}
	$load=~s:.*[\\/](.*$):$1:o; # strip off the UNIX or DOS filename
	$dst="$dir$load";
	print STDERR "src=$load dst=$dst\n" if $debug;
	unless(($count=Upload($r,$cfg,$docroot,$load,$dst))) {
	    $r->log->warn(__PACKAGE__." $cmd ERROR: $user: $proc src=$load dst=$dst");
	    $retval=$error;
	} else {
	    $msg="$user: $cmd: $count: $load -> $dst";
	}
	$r->log->notice(__PACKAGE__." $msg");
	return $retval;
    }
    my $icnt=@$items;
    my $select=$icnt;
    $formsrc="$dir$items->[--$icnt]" if $select;
    unless(dirbound($formsrc,$fmroot)) { # Don't allow $formsrc below fmroot
	$msg=$lang->{SourcePath} || 'Bad source path';
	errmsg($msg);
	return $error;
    } 
    if($cmd eq 'View') {
        $formsrc=~s:/$rootdir/:/: if $proc eq 'root';
	$msg="$user: $cmd: $formsrc";
	$r->log->notice(__PACKAGE__." $msg");
	$r->header_out(Location=>$formsrc);
	return REDIRECT;
    }
    if($cmd eq 'Revoke') {
	$formsrc="$docroot$fmroot$rootdir".REVOKE_DIR;
	unless(-e $formsrc) {
	    chmod 0750,"$docroot$fmroot$rootdir";	# Attempt to create file
	    unless(mkdir $formsrc,0750) {		# If it does not exist
		$msg=$lang->{create} || 'Can\'t create file';
		$msg.=" $fmroot".REVOKE_DIR;
		$msg.=" $!";
		errmsg($msg);
		chmod 0550,"$docroot$fmroot$rootdir";
		return $error;
	    }
	    chmod 0550,"$docroot$fmroot$rootdir";
	}
	unless(($retval=Revoke($r,$cfg))) {
	    $r->log->warn(__PACKAGE__." $cmd ERROR: $user: $errmsg $formsrc");
	    $retval=$error;
	} else {
	    $msg="$user: $cmd: $formsrc";
	    $r->log->notice(__PACKAGE__." $msg");
	}
	return $retval;
    }
    unless(dirbound($formdst,$fmroot)) { # Don't allow $formdst below fmroot
	$msg=$lang->{ProcDstRoot} || 'Destination goes below the root directory';
	errmsg($msg);
	return $error;
    } 
    my $oldmask=umask $umask if $umask;
    if($cmd eq 'MkDir') {
	print STDERR "dst=$formdst\n" if $debug;
	unless(MkDir($r,$cfg,$docroot,$formsrc,$formdst)) {
	    $r->log->warn(__PACKAGE__." $cmd ERROR: $user: $errmsg $proc dst=$formdst");
	    $retval=$error;
	}
	$msg="$user: $cmd: $formdst";
	$r->log->notice(__PACKAGE__." $msg");
	umask($oldmask) if $umask;
	return $retval;
    }
    if($cmd eq 'MkFile') {
	$formsrc=$formdst;
	print STDERR "dst=$formdst\n" if $debug;
	unless(Edit($r,$cfg,$docroot,$formsrc)) {
	    $r->log->warn(__PACKAGE__." $cmd ERROR: $user: $errmsg $proc src=$formsrc");
	    $retval=$error;
	}
	$msg="$user: $cmd: $formsrc";
	$r->log->notice(__PACKAGE__." $msg");
	umask($oldmask) if $umask;
	return $retval;
    }
    if($formsrc eq $fmroot || ($cfg->{options}&FANCY_INDEXING && !$select)) {
	$msg=$lang->{ProcSelect} || 'Nothing selected!';
	errmsg($msg);
	return $error;
    }
# Command can only have only one source item.
    if($icnt && $cmd eq 'Rename') {
	$msg=$lang->{ProcTooMany} || 'More than one item selected';
	errmsg($msg);
	$r->log->warn(__PACKAGE__." $cmd ERROR: $user: $errmsg $proc src=$formsrc");
	umask($oldmask) if $umask;
	return $error;
    }
    do {
	if($cmd eq 'SetGID') {
	    $formdst="$args->{gid}";
	    print STDERR "src=$formsrc gid=$formdst\n" if $debug;
	    $msg="$user: $cmd: $formdst $formsrc";
	} elsif($cmd eq "Delete") {
	    print STDERR "src=$formsrc\n" if $debug;
	    $msg="$user: $cmd: $formsrc";
	} else {
	    print STDERR "src=$formsrc dst=$formdst\n" if $debug;
	    $msg="$user: $cmd: $formsrc -> $formdst";
	}
	no strict 'refs';
	my $routine=\&$cmd;
	use strict 'refs';
	$retval=&{$routine}($r,$cfg,$docroot,$formsrc,$formdst);
	unless($retval) {
	    $r->log->warn(__PACKAGE__." $cmd ERROR: $user: $errmsg $proc docroot=$docroot src=$formsrc dst=$formdst");
	    $retval=$error;
	} else {
	    $msg="$user: $cmd: $formsrc -> $formdst";
	    $r->log->notice(__PACKAGE__." $msg");
	    $retval=0 if $retval>0;
	}
	if(!$select && !$retval && ($cmd eq 'Rename' || $cmd eq 'Move' || $cmd eq 'Delete')) {
	    $dst=$redir;
	    if($proc eq 'hold') {
		$dst.="?root_dir=$args->{root_dir}";
		$args->{hold_file}='';
		$args->{hold_dst}='';
	    } else {
		$dst.="?hold_dir=$args->{hold_dir}" if $hold;
		$args->{root_file}='';
		$args->{root_dst}='';
	    }
	    $r->log->notice(__PACKAGE__." $msg");
	    $r->header_out(Location=>"$dst#$proc");
    	    umask($oldmask) if $umask;
	    return REDIRECT;
	}
	if($icnt) {
	    $formsrc="$dir$items->[--$icnt]";
	} else {
	    $retval=1 unless $retval;
	}
    } until $retval;
    umask($oldmask) if $umask;
    $retval;
}

sub getcmd {
    my ($cmds, $args)=@_;
    foreach (@$cmds) {
	return $_ if $args->{$_};
    }
    '';
}

sub isadmin {
    my ($check)=@_;
    return 0 unless $check;	# don't allow root gid
    return 1 unless @$gid;	# always a member if no gid 
    for(my $cnt=@$gid-1;$cnt>=0;$cnt--) {
	return 1 if $gid->[$cnt]==$check;
    }
    0;
}

sub chgid {
    return 1 unless @$gid;
	chown(-1,$chgid,$_[1]);
    1;
}

sub SetGID {	# Set the item (file or dir) GID 
    my ($r,$cfg,$root,$src,$igid) = @_;
    $src="$root$src";
    my $name;
    my $lang = new Apache::Language($r) if $cfg->{language};
    my $msg='';
    my $cmdname=$lang->{SetGID} || 'SetGID';
    if(isadmin($admin)) {
	if($igid=~m:[^0-9]:o) {		# if not a number look-up the group
	    $name=$igid;
	    unless(($igid=getgrnam $name)) {
		$msg=$lang->{GIDbad} || 'GID not found';
		errmsg(qq~${cmdname}: "$name" $msg~);
		return 0;
	    }
	} else {
	    unless(($name=getgrgid $igid)) {
		$msg=$lang->{GIDbad} || 'GID not found';
		errmsg(qq~${cmdname}: "$igid" $msg~);
		return 0;
	    }
	}
	unless($igid && chown(-1,$igid,$src)) {
	    $msg=$lang->{GIDset} || 'GID not set';
	    errmsg(qq~${cmdname}: "$name" $msg~);
	    return 0;
	}
    } else {
	$msg=$lang->{internal} || 'internal';
	errmsg("${cmdname}: $msg");
	return 0;
    }
}

sub Revoke {
    my ($r,$cfg) = @_;
    my $uri = $r->uri;
    my $textlen=$cfg->{textlen} || DEFAULT_TEXT_LEN;
    my $halflen=($textlen+($textlen%2))/2;
    my $lang = new Apache::Language($r) if $cfg->{language};
    my $msg='';
    my $cmdname=$lang->{Revoke} || 'Revoke';
    if(!$cfg->{revoke} || !isadmin($admin)) {
	$r->log->error(__PACKAGE__." Revoke: internal error:");
	$msg=$lang->{internal} || 'internal';
	errmsg("${cmdname}: $msg");
	return 0;
    }
    $msg=$lang->{FileManRevoke} || 'FileMan Revoke';
    return SKIP_INDEX unless header($r,$msg); 
    my $gotdata;
    my $type;
    my $name;
	foreach (keys %$Users) {
	    if($Users->{$_} eq '-') {
		my($ruser,$rgid)=m:^(.*?)#(.*?)#:;
		unless($gotdata) {
		    $msg=$lang->{Revoked} || 'The following have been revoked:';
		    print "$msg<P>\n";
		    print qq~<TABLE COL="2" BORDER="0"><TR>\n~;
    	    print qq~<TH> Type </TH><TH> Name </TH><TR>\n~;
		    $gotdata=1;
		}
		if($ruser) {
		    $type='user';
		    $name=$ruser;
		}
		if($rgid) {
		    $type='gid';
		    $name=$rgid;
		}
		print "<TD> $type </TD><TD> $name </TD><TR>\n";
	    }
	}
	print "</TABLE>\n" if $gotdata;
	unless($gotdata) {
	    $msg=$lang->{NoUsers} ||
	        'No user or group revoke information available';
	    print "$msg<P>\n";
	}
	print qq~<FORM METHOD="POST" ACTION="$uri" ENCTYPE="MULTIPART/FORM-DATA">\n~;
	print qq~<INPUT TYPE="TEXT" NAME="id" SIZE=$halflen MAXLENGTH=255>\n~;
	  $msg=$lang->{EnableUID} || 'Enable User';
	  chomp $msg;
	print qq~<INPUT TYPE="SUBMIT" NAME="enauid" VALUE="$msg">~;
	  $msg=$lang->{DisableUID} || 'Disable User';
	  chomp $msg;
	print qq~<INPUT TYPE="SUBMIT" NAME="disuid" VALUE="$msg">\n~;
	  $msg=$lang->{EnableGID} || 'Enable GID';
	  chomp $msg;
	print qq~<INPUT TYPE="SUBMIT" NAME="enagid" VALUE="$msg">~;
	  $msg=$lang->{DisableGID} || 'Disable GID';
	  chomp $msg;
	print qq~<INPUT TYPE="SUBMIT" NAME="disgid" VALUE="$msg"><P>\n~;
	  $msg=$lang->{Return} || 'Return';
	  chomp $msg;
	print qq~<INPUT TYPE="SUBMIT" NAME="return" VALUE="$msg">~,
	      qq~<INPUT TYPE="HIDDEN" NAME="proc" VALUE="revoke"></FORM><HR>\n~;
    SKIP_INDEX;
}

sub Edit {
    my ($r,$cfg,$root,$src) = @_;
    my $relsrc=$src;
    $src="$root$src";
    my $lang = new Apache::Language($r) if $cfg->{language};
    my $msg='';
    my $cmdname=$lang->{Edit} || 'Edit';
    my %info;
    my $inifile;
    my $opened;
    my $uri = $r->uri;
    my $fgid=(stat $src)[5];
    if(-e _) {
	unless(isadmin($fgid) || isadmin($admin)) {
	    $msg=$lang->{SourceAccess} || 'Source access denied';
	    errmsg("${cmdname}: $msg");
	    return 0;
	}
	unless(-f _) {
	    $msg=$lang->{NotText} || 'Item is not a text file';
	    errmsg("${cmdname}: $msg");
	    return 0;
	}
	unless(-T _) {
	    $msg=$lang->{NotText} || 'Item is not a text file';
	    errmsg("${cmdname}: $msg");
	    return 0;
	}
	my $editmax=$cfg->{editmax} | DEFAULT_EDIT_MAX;
	unless(-s _ <= $editmax) {
	    $msg=$lang->{FileTooBig} || 'File size is larger than';
	    errmsg("${cmdname}: $msg $editmax");
	    return 0;
	}
	unless(open ITEM, "<$src") {
	    $msg=$lang->{FileOpen} || 'File open';
	    errmsg("${cmdname}: $msg");
	    return 0;
	}
	$opened=1;
    } else {
	my ($parent)=$src=~m:(^.*)/.+:o;
	my $fgid=(stat $parent)[5];
	unless(isadmin($fgid) || isadmin($admin)) {
	    $msg=$lang->{ParentAccess} || 'Parent access denied';
	    errmsg("${cmdname}: $msg");
	    return 0;
	}
    }
    ($inifile=$src)=~s:^(.*/)(.+):$1\.$2\.ini:;
    if(open INIFILE,"<$inifile") {
	$info{open}=1;
	while(<INIFILE>) {
	    chomp;
	    my($key,$value)=m:(\w+)\s*=\s*(.+):;
	    $info{$key}=$value;
	}
	close INIFILE;
    }
    $msg=$lang->{FileManEdit} || 'FileMan Edit';
    return SKIP_INDEX unless header($r,"$msg $relsrc"); 
    if($info{status} eq 'out') {
	$msg=$lang->{warning} || 'WARNING';
	$errmsg="${msg}:";
	if($user) {
	    $msg=$lang->{User} || 'User';
	    $errmsg.=qq~ $msg "$info{user}"~;
	}
	$msg=$lang->{CheckedOut} ||
	  'Currently has checked out';
	$errmsg.=qq~ $msg "$relsrc"~;
	$r->log->warn(__PACKAGE__." Edit: $errmsg");
	print "<H3><FONT COLOR=#FF0000>$errmsg</FONT></H3>\n";
    } else {
	$msg=$lang->{Editing} || 'Editing';
	print qq~<H3>$msg "$relsrc"</H3>\n~;
    }
    unless(open INIFILE, ">$inifile") {
	$msg=$lang->{FileLocked} || 'File locked';
	errmsg("${cmdname}: $msg");
	return 0;
    }
    unless(flock INIFILE, LOCK_EX|LOCK_NB) {
	$msg=$lang->{NotLocked} || 'Could not lock file. Try again.';
	errmsg("${cmdname}: $msg");
	return 0;
    }
    print INIFILE "edited=$info{editedby}\ngid=$info{gid}\ntime=$info{time}\nuser=$user\nstatus=out\n";
    flock INIFILE, LOCK_UN;
    close INIFILE;
    if($info{open}) {
	$msg=$lang->{EditLast} || 'Last edit information:';
	chomp $msg;
	my $phrase=$msg;
	$msg=$lang->{User} || 'User';
	chomp $msg;
	$phrase.=" $msg";
	$phrase.="=$info{editedby}" if $info{editedby};
	$msg=$lang->{Access} || 'Access';
	chomp $msg;
	$phrase.=" $msg";
	$phrase.="=$info{gid}" if $info{gid};
	$msg=$lang->{Time} || 'Time';
	chomp $msg;
	$phrase.=" ${msg}=$info{time}";
	print "$phrase<P>\n";
    }
    print qq~<FORM METHOD="POST" ACTION="$uri" ENCTYPE="MULTIPART/FORM-DATA">\n~;
	$msg=$lang->{Undo} || 'Undo';
	chomp $msg;
    print qq~<INPUT TYPE="RESET" NAME="undo" VALUE="$msg">\n~;
	$msg=$lang->{Quit} || 'Quit';
	chomp $msg;
    print qq~<INPUT TYPE="SUBMIT" NAME="quit" VALUE="$msg">\n~;
	$msg=$lang->{Save} || 'Save';
	chomp $msg;
    print qq~<INPUT TYPE="SUBMIT" NAME="save" VALUE="$msg">\n~,
    	  qq~<TEXTAREA NAME="text" ROWS="24" COLS="80" WRAP="physical">\n~;
    if($opened) {
	while(<ITEM>) {
	    chomp;
	    print(escape_html($_));
	}
	close ITEM;
    }
    ($inifile=$relsrc)=~s:^(.*/)(.+):$1\.$2\.ini:;
    print "</TEXTAREA><P>\n",
	qq~<INPUT TYPE="HIDDEN" NAME="proc" VALUE="edit">\n~,
	qq~<INPUT TYPE="HIDDEN" NAME="file" VALUE="$relsrc">\n~,
	qq~<INPUT TYPE="HIDDEN" NAME="info" VALUE="$inifile">\n~,
	qq~<INPUT TYPE="HIDDEN" NAME="user" VALUE="$user">\n~;
    print qq~</FORM>\n~;
    if($debug && $dodump) {
	use Data::Dumper;
	print "<HR><PRE>";
	print "\%info\n";
	print Dumper \%info;
	print "</PRE><HR>";
    }
    SKIP_INDEX;
}

sub EditSave {
    my ($r,$cfg,$docroot,$args)=@_;
    my $file="$docroot$args->{file}";
    my $lang = new Apache::Language($r) if $cfg->{language};
    my $msg='';
    my $cmdname=$lang->{MkDir} || 'MkDir';
    unless(open FILE, ">$file") {
	$msg=$lang->{FileOpen} || 'File Open';
	errmsg("${cmdname}: $msg");
	return 0;
    }
    print FILE $args->{text};
    close FILE;
	my ($parent)=$file=~m:(^.*)/.+:o;
	my $fgid=(stat $parent)[5];
	chown(-1,$fgid,$file);
    1;
}

sub revoke {
    my ($r,$cfg,$type,$action,$name,$file)=@_;
    my $retval=0;
    my $curval;
    my $lang = new Apache::Language($r) if $cfg->{language};
    my $msg='';
    my $cmdname=$lang->{Revoke} || 'Revoke';
    unless($name) {
	$msg=$lang->{RevokeName} || 'No ID number or name provided';
	errmsg("${cmdname}: $msg");
	return 0;
    }
    $name= lc $name;
    if(-e "$file.new") {	# gross file locking, should never happen
	$r->warn(__PACKAGE__ . " revoke file locked: ${file}.new exists");
	$msg=$lang->{FileLocked} || 'File locked';
	errmsg("${cmdname}: $msg");
	return 0;
    }
    my $server=$r->get_server_name;
    $curval=$Users->{"#${name}#${server}#${fmroot}"} if $type eq 'gid';
    $curval=$Users->{"${name}##${server}#${fmroot}"} if $type eq 'user';
    if($action eq '-') {	# '-' implies disable user/group
	return 1 if $curval eq '-';
	if(open REVOKE, ">>$file") {	# append name to the revoke file
	    print REVOKE "$type=$name\n";
	    close REVOKE;
	    $retval=1;
	} else {
	    $r->warn(__PACKAGE__ . " revoke file append open");
	    $msg=$lang->{FileOpen} || 'File open';
	    errmsg("${cmdname}: $msg");
	}
    } elsif($action eq '+') {	# '+' implies enable user/group
	return 1 if $curval eq '+';
	if(open REVOKE, "<$file") {	# remove name from revoke file 
	    if(open NEWREVOKE, ">$file.new") {
		my($key,$val);
		while(<REVOKE>) {
		    ($key,$val)=m:(\w+)\s*=\s*(\w+):;
		    print NEWREVOKE "$key=$val\n" if $key && $key ne $type && $val ne $name;
		}
		close NEWREVOKE;
		close REVOKE;
		unlink $file;
		rename "$file.new", $file;
		$retval=1;
	    } else {
		$r->warn(__PACKAGE__ . " revoke file write open");
		$msg=$lang->{FileOpen} || 'File open';
		errmsg("${cmdname}: $msg");
		close REVOKE;
	    }
	}
    }
    if($retval) {
	$Users->{"#${name}#${server}#${fmroot}"}=$action if $type eq 'gid';
	$Users->{"${name}##${server}#${fmroot}"}=$action if $type eq 'user';
    }
    $retval;
}

# Internal command methods
sub MkDir {
    my ($r,$cfg,$root,$src,$dst) = @_;
    my $lang = new Apache::Language($r) if $cfg->{language};
    my $msg='';
    my $cmdname=$lang->{MkDir} || 'MkDir';
    $dst="$root$dst";
    if(-e $dst) {
	$msg=$lang->{DestExists} || 'Destination exists';
	errmsg("${cmdname}: $msg");
	return 0;
    }
    if(@$gid) {
	my ($parent)=$dst=~m:(^.*)/.+:o;
	my $fgid=(stat $parent)[5];
	unless(isadmin($fgid) || isadmin($admin)) {
	    $msg=$lang->{ParentAccess} || 'Parent access denied';
	    errmsg("${cmdname}: $msg");
	    return 0;
	}
	mkdir $dst,0755;
	chown(-1,$fgid,$dst);
    } else {
	mkdir $dst,0755;
    }
}

sub Unzip {
    my ($r,$cfg,$root,$src,$dst) = @_;
    $dst=~s:/$::;		# strip any trailing '/'
    my $target=$dst;
    use Archive::Zip qw( :ERROR_CODES :CONSTANTS );
    use Archive::Zip::Tree;
    my $lang = new Apache::Language($r) if $cfg->{language};
    my $msg='';
    my $cmdname=$lang->{Unzip} || 'Unzip';
    unless($dst) {
	$msg=$lang->{DestPath} || 'Bad destination path';
	errmsg("${cmdname}: $msg");
	return 0;
    }
    $src="$root$src";
    $dst="$root$dst";
    my $fgid=(stat $src)[5];
    unless(isadmin($fgid) || isadmin($admin)) {
	$msg=$lang->{SourceAccess} || 'Source access denied';
	errmsg("${cmdname}: $msg");
	return 0;
    }
    $fgid=(stat $dst)[5];
    if(! -d _) {
	$msg=$lang->{DestDir} || 'Destination is not a directory';
	errmsg("${cmdname}: $msg");
	return 0;
    }
    unless(isadmin($fgid) || isadmin($admin)) {
	$msg=$lang->{DestAccess} || 'Destination access denied';
	errmsg("${cmdname}: $msg");
	return 0;
    }
    my $zip=Archive::Zip->new($src);
    unless ($zip) {
	$msg=$lang->{FileRead} || 'file read';
	errmsg("${cmdname}: $msg");
	return 0;
    }
    my $files=0;
    my $name;
    $dst.='/';
    for my $member ($zip->members()) {
	($name=$dst).=$member->fileName();
	if($member->isDirectory()) {
	    mkdir $name,0775;
	    chown(-1,$fgid,$name);
            next;
	}
	unless($member->extractToFileNamed($name)==AZ_OK) {
	    errmsg("$cmdname: $name");
	    return 0;
	}
	chown(-1,$fgid,$name);
	++$files;
    }
    $files;
}

sub Move {
    my ($r,$cfg,$root,$src,$dst) = @_;
    my $target=$src;
    $src="$root$src";
    $dst="$root$dst";
    use File::Copy qw(move);
    my $lang = new Apache::Language($r) if $cfg->{language};
    my $msg='';
    my $cmdname=$lang->{Move} || 'Move';
    unless($target) {
	$msg=$lang->{DestPath} || 'Bad destination path';
	errmsg("${cmdname}: $msg");
	return 0;
    }
    $target=~s:^.*/(.*):$1:;
    my $srcgid=(stat $src)[5];
    my $src_is_dir=1 if -d _;
	unless(isadmin($srcgid) || isadmin($admin)) {
	$msg=$lang->{SourceAccess} || 'Source access denied';
	errmsg("${cmdname}: $msg");
	return 0;
    }
    my $dstgid=(stat $dst)[5];
	unless(isadmin($dstgid) || isadmin($admin)) {
	$msg=$lang->{DestAccess} || 'Destination access denied';
	errmsg("${cmdname}: $msg");
	return 0;
    }
    $dst.=$target if $src_is_dir;
    unless(File::Copy::move($src, $dst)) {
	errmsg("${cmdname}: $!");
	return 0;
    }
	chown(-1,$dstgid,$dst) unless isadmin($admin); # admin can move others
    return 1;
}

sub Rename {
    my ($r,$cfg,$root,$src,$dst) = @_;
    my $target=$dst;
    $src="$root$src";
    $dst="$root$dst";
    use File::Copy qw(move);
    my $lang = new Apache::Language($r) if $cfg->{language};
    my $msg='';
    my $cmdname=$lang->{Rename} || 'Rename';
    my $fgid=(stat $src)[5];
    unless(isadmin($fgid) || isadmin($admin)) {
	$msg=$lang->{SourceAccess} || 'source access denied';
	errmsg("${cmdname}: $msg");
	return 0;
    }
    unless($target) {
	$msg=$lang->{DestPath} || 'bad destination path';
	errmsg("${cmdname}: $msg");
	return 0;
    }
    if(-e $dst) {
	$msg=$lang->{DestExists} || 'Destination';
	errmsg("${cmdname}: $msg");
	return 0;
    }
    unless(File::Copy::move($src, $dst)) {
	errmsg("${cmdname}: $!");
	return 0;
    }
    return 1;
}

sub Copy {
    my ($r,$cfg,$root,$src,$dst) = @_;
    my $target=$src;
    $src="$root$src";
    $dst="$root$dst";
    use File::NCopy qw(copy);
    my $lang = new Apache::Language($r) if $cfg->{language};
    my $msg='';
    my $cmdname=$lang->{Copy} || 'Copy';
    unless($target) {
	$msg=$lang->{DestPath} || 'Bad destination path';
	errmsg("${cmdname}: $msg");
	return 0;
    }
    my $file;
    my $isdir;
    my $fgid=(stat $dst)[5];
	$chgid=0;
    if(-e _) {
    $isdir=1 if -d _;
	unless(isadmin($fgid) || isadmin($admin)) {
	    $msg=$lang->{DestAccess} || 'Destination access denied';
	    errmsg("${cmdname}: $msg");
	    return 0;
	}
	$chgid=$fgid;	# global used by chgid method to set GID
    }
    $fgid=(stat $src)[5];
    if(! -e _) {
	$msg=$lang->{SourcePath} || 'Bad source path';
	errmsg("${cmdname}: $msg");
	return 0;
    }
    unless(isadmin($fgid) || isadmin($admin)) {
	$msg=$lang->{SourceAccess} || 'Source access denied';
	errmsg("${cmdname}: $msg");
	return 0;
    }
    $chgid=$fgid unless $chgid;
    if(-d _) {
	unless($isdir) {
	    $msg=$lang->{DirConflict} || 'Source directory but a destination file';
	    errmsg("${cmdname}: $msg");
	    return 0;
	}
	if($dst=~m:^$src:) {
	    $msg=$lang->{CopyRecusive} || 'Recursive copy detected';
	    errmsg("${cmdname}: $msg");
	    return 0;
	}
	$file=File::NCopy->new
	(recursive=>1,force_write=>1,set_permission=>\&chgid);
    } else {
	$file=File::NCopy->new
	(force_write=>1,set_permission=>\&chgid);
    }
    return 1 if $file->copy($src, $dst);
    $msg=$lang->{DestCheck} || 'Check destination path';
    errmsg("${cmdname}: $msg: $!");
    return 0;
}

sub Delete {
    my ($r,$cfg,$root,$src) = @_;
    $src="$root$src";
    use File::Path qw(rmtree);
    my $lang = new Apache::Language($r) if $cfg->{language};
    my $msg='';
    my $cmdname=$lang->{Delete} || 'Delete';
    my $fgid=(stat $src)[5];
    unless(-e _) {
	$msg=$lang->{SourcePath} || 'Bad source path';
	errmsg("${cmdname}: $msg");
	return 0;
    }
    unless(isadmin($fgid) || isadmin($admin)) {
	$msg=$lang->{SourceAccess} || 'Source access denied';
	errmsg("${cmdname}: $msg");
	return 0;
    }
    if(-d _) {
	unless(File::Path::rmtree($src)) {
	    errmsg("${cmdname}: $!");
	    return 0;
	}
    } else {
	unless(unlink($src)) {
	    errmsg("${cmdname}: $!");
	    return 0;
	}
    }
}

sub Upload {
    my ($r,$cfg,$root,$src,$dst) = @_;
    $dst="$root$dst";
    my $upload=$r->upload;
    my $sfh=$upload->fh;
    my $bytes=0;
    my $size=0;
    my $lang = new Apache::Language($r) if $cfg->{language};
    my $msg='';
    my $cmdname=$lang->{Upload} || 'Upload';
    unless($sfh) {
	$msg=$lang->{internal} || 'internal';
	errmsg("${cmdname}: $msg");
	return 0;
    }
    my ($parent)=$src=~m:(^.*)/.+:o;
	my $fgid=(stat $parent)[5];
	unless(isadmin($fgid) || isadmin($admin)) {
	$msg=$lang->{ParentAccess} || 'Parent access denied';
	errmsg("${cmdname}: $msg");
	return 0;
	}
    unless(open DFH, ">$dst") {
	$msg=$lang->{DestOpen} || 'Destination open';
	errmsg("${cmdname}: $msg");
	return 0;
    }
    my $buf;
    while(($size=read($sfh, $buf, 4096))) {
	unless(print DFH $buf) {
	    close DFH;
	    $msg=$lang->{write} || 'write';
	    errmsg("${cmdname}: $msg");
	    return 0;
	}
	$bytes+=$size;
    }
    close DFH;
    chown(-1,$fgid,"$root$dst");
    return $bytes;
}
# End of Internal command methods

sub errmsg {
    return $errmsg unless defined $_[0];
    ($errmsg)=shift;
}

sub hold_form {
    my ($r,$args,$cfg,$mode,$menu) = @_;
    if($args->{hold_err}) {
	print "<H3><FONT COLOR=#FF0000>ERROR: ",errmsg(),"</FONT></H3>\n";
	$args->{hold_err}=0;
    }
    cmd_form($r,'hold',$args,$cfg,$mode,$menu);
}

sub root_form {
    my ($r,$args,$cfg,$mode,$menu) = @_;
    if($args->{root_err}) {
	print "<H3><FONT COLOR=#FF0000>ERROR: ",errmsg(),"</FONT></H3>\n"; 
	$args->{root_err}=0;
    }
    cmd_form($r,'root',$args,$cfg,$mode,$menu);
}

sub cmd_form {
    my ($r,$name,$args,$cfg,$mode,$menu)=@_;
    my $uri=$r->uri;
    my $src;
    my $dst;
    my $setgid;
    my $docroot=$r->document_root;
    my $rootdir=$cfg->{rootdir};
    my $textlen=$cfg->{textlen} || DEFAULT_TEXT_LEN;
    if($name eq 'hold') {
	if(!$args->{hold_file} && $args->{hold_child}) {
	    $src=$args->{hold_child};
	} else {
	    $src=$args->{hold_file};
	}
	$dst=$args->{hold_dst};
	($dst=$args->{root_dir})=~s:/$rootdir/:/: unless $dst;
	$setgid=$args->{hold_gid};
    } elsif ($name eq 'root') {
	if(!$args->{root_file} && $args->{root_child}) {
	    $src=$args->{root_child};
	} else {
	    $src=$args->{root_file};
	}
	$dst=$args->{root_dst};
	$dst=$args->{hold_dir} unless $dst || !$hold;
	$setgid=$args->{hold_gid};
    } else {
	errmsg("unknown internal form");
	print "<H3><FONT COLOR=#FF0000>ERROR: ",errmsg(),"</FONT></H3>\n"; 
	return 0;
    }
    my $didit;
    my $msg='';
    my $lang = new Apache::Language($r) if $cfg->{language};
    foreach (@$menu) {
	if($_ eq 'Upload') {
	    $msg=$lang->{$_} || $_;
	    chomp $msg;
	    print qq~<INPUT TYPE="FILE" NAME="browse" SIZE=$textlen MAXLENGTH=255>\n~,
		  qq~<INPUT TYPE="SUBMIT" NAME="$_" VALUE="$msg"><P>\n~;
	}
    }
    foreach (@$menu) {
	unless($_ eq 'Upload') {
	    $msg=$lang->{$_} || $_;
	    chomp $msg;
	    print qq~<INPUT TYPE="SUBMIT" NAME="$_" VALUE="$msg">\n~;
	}
    }
    unless($mode) {
	$msg=$lang->{Source} || 'Select Item';
	chomp $msg;
	print qq~<P><INPUT TYPE="TEXT" NAME="${name}_src" SIZE=$textlen MAXLENGTH=255 VALUE="$src">$msg\n~;
    }
    $msg=$lang->{Destination} || 'Destination';
    chomp $msg;
    print qq~<P><INPUT TYPE="TEXT" NAME="${name}_dst" SIZE=$textlen MAXLENGTH=255 VALUE="$dst">$msg<P>\n~;
    if(isadmin($admin)) {
	my $halflen=($textlen+($textlen%2))/2;
	  $msg=$lang->{SetGID} || 'SetGID';
	  chomp $msg;
	print qq~<INPUT TYPE="TEXT" NAME="gid" SIZE=$halflen MAXLENGTH=255>$setgid\n~,
	      qq~<INPUT TYPE="SUBMIT" NAME="SetGID" VALUE="$msg">\n~;
	  $msg=$lang->{Revoke} || 'Revoke';
	  chomp $msg;
	print qq~<INPUT TYPE="SUBMIT" NAME="Revoke" VALUE="$msg">\n~ if $cfg->{revoke};
	  $msg=$lang->{Debug} || 'Debug';
	  chomp $msg;
	print qq~<INPUT TYPE="SUBMIT" NAME="Debug" VALUE="$msg">\n~ if $debug;
	print qq~<P>\n~;
    }
    print qq~<INPUT TYPE="HIDDEN" NAME="proc" VALUE="${name}">\n~;
    foreach (keys %$args) {
	print qq~<INPUT TYPE="HIDDEN" NAME="$_" VALUE="$args->{$_}">\n~ unless m:^$name:o || !m:_:o;
    }
    1;
}

sub plain_page {
    my ($r,$args,$dirhandle,$tree,$isroot)=@_;
    print "<UL>\n";
    while (my $file = readdir $dirhandle) {
	my $stub;
	next if $file eq '.';
	next if $file eq ".." and $isroot;
	my $subr = $r->lookup_file($file);
	stat $subr->finfo;
	print '    <LI><A HREF="',$args->{"${tree}_dir"},$file;
	print '/' if -d _;
	if($tree eq 'root') {
	    if($file eq '..') {
		$stub=$args->{root_dir};
		$stub=~s:/$::;
		$stub=~s:.*/::;
	    }
	    if($hold) {
		print "?hold_dir=$args->{hold_dir}&hold_file=$args->{hold_file}";
		print "&root_child=" if $stub;
	    } else {
		print "?root_child=" if $stub;
	    }
	    print $stub if $stub;
	    print '#root';
	} elsif($tree eq 'hold') {
	    if($file eq '..') {
		$stub=$args->{hold_dir};
		$stub=~s:/$::;
		$stub=~s:.*/::;
	    }
	    print "?root_dir=$args->{root_dir}&root_file=$args->{root_file}";
	    print "&hold_child=$stub" if $stub;
	    print '#hold';
	}
    	if($file eq $args->{"${tree}_file"}) {
	    print qq~"><B>$file</B></A></LI>\n~;
    	} else {
	    print qq~">$file</A></LI>\n~;
    	}
    }
    print "</UL><HR>\n";
    return OK;
}

sub fancy_page {
    my ($r,$args,$dirhandle,$mode,$isroot)=@_;
    my $msg='';
    my $cfg  = Apache::ModuleConfig->get($r);
    my $subr;
    my $uri = $r->uri;
    my $isadmin=@$gid && isadmin($admin);
    my $lang = new Apache::Language($r) if $cfg->{language};
    my $table_html = $r->dir_config("IndexHtmlTable") || 'BORDER="0"';
    print "<TABLE $table_html>\n<TR>";
    my $list = read_dir( $r, $dirhandle);
    if ($cfg->{options} & SUPPRESS_COLSORT) {
	foreach ('N', 'M', 'S', 'D') {
	    delete $args->{@_};
	}
    }
    my $listing = do_sort($list,$args,$cfg->{default_order});
#Perms header
    print '<TH ALIGN="LEFT">Perms</TH>' if $cfg->{options} & SHOW_PERMS;
#Owner header
    print '<TH ALIGN="LEFT">Owner</TH>' if $isadmin;
#Group header
    $msg=$isadmin?"Group":"Access";

    print qq~<TH ALIGN="LEFT">$msg</TH>\n~ if @$gid;
#Select header
    print '<TH ALIGN="CENTER">Select</TH>' if $mode;
#Icon header
    print '<TH ALIGN="LEFT">Icon</TH>' if $cfg->{options} & FANCY_INDEXING;
#Name, Last Modified, Size, and Description headers
    foreach ('N', 'M', 'S', 'D') {
	next if $cfg->{options} & SUPPRESS_LAST_MOD && $_ eq 'M';
	next if $cfg->{options} & SUPPRESS_SIZE     && $_ eq 'S';
	next if $cfg->{options} & SUPPRESS_DESC     && $_ eq 'D';
	print '<TH ALIGN="LEFT">';
	$msg=$lang->{$sortname{$_}} || $sortname{$_};
	if(not $cfg->{options} & SUPPRESS_COLSORT) {
	    if($args->{$_}) {
		my $query;
		if($_ eq 'N') {		# Name, can sort on extention
		    $query=($args->{$_} eq 'D')?'A':($args->{$_} eq 'A')?'E':'D';
		} else {
		    $query = ($args->{$_} eq 'D')?'A':'D';
		}
		print qq~<A HREF="?$_=$query"><I>$msg</I></A>~;
	    } else {
		print qq~<A HREF="?$_=D">$msg</A>~;
	    }
	} else {
	    print $msg;
    	}
        print "</TH>\n";
    }
    print "</TR>";
#End of header
    for my $entry (@$listing) {
	my $stub;
	my $label='';
	if($entry eq '..') {
	    next if $isroot;
	    $label=$lang->{Parent} || 'Parent&nbsp;Directory';
	} else {
	    $label = $entry;
	}
	my $img = $list->{$entry}{icon};
	print qq~<TR ALIGN="LEFT">~;
#Perms data
    print qq~<TD>$list->{$entry}{mode}</TD>\n~ if $cfg->{options} & SHOW_PERMS;
#Owner data
    if($isadmin) {
	my $pname=getpwuid($list->{$entry}{uid})||"$list->{$entry}{uid}";
	print "<TD>${pname}</TD>\n";
    }
#Group data
    if(@$gid) {
	my $pname=getgrgid($list->{$entry}{gid})||"$list->{$entry}{gid}";
	print "<TD>${pname}</TD>\n";
    }
	if($mode eq 'hold') {
	    if($entry eq '..') {
		$stub=$args->{hold_dir};
		$stub=~s:/$::;
		$stub=~s:.*/::;
	    }
	} elsif($mode eq 'root') {
	    if($entry eq '..') {
		$stub=$args->{root_dir};
		$stub=~s:/$::;
		$stub=~s:.*/::;
	    }
	}
	my $isdir=1 if $list->{$entry}{sizenice} eq '-';
#Select checkbox
	print qq~<TD ALIGN="CENTER"><INPUT TYPE="CHECKBOX" NAME="${entry}"></TD>\n~ if $mode;
#Icon
	print '<TD>';
	if ($cfg->{options} & ICONS_ARE_LINKS) {
	    $msg=$args->{"${mode}_dir"};
	    $msg=~s:/$cfg->{rootdir}/:/: if $mode eq 'root' && !$isdir;
	    print qq~<A HREF="$msg$entry~;
	    print '/' if $isdir;
	    if($mode eq 'root') {
		if($hold) {
		    print "?hold_dir=$args->{hold_dir}&hold_file=$args->{hold_file}";
		    print "&root_child=" if $stub;
		} else {
		    print "?root_child=" if $stub;
		}
		print $stub if $stub;
		print '#root';
	    } elsif($mode eq 'hold') {
		print "?root_dir=$args->{root_dir}&root_file=$args->{root_file}";
		print "&hold_child=$stub" if $stub;
		print '#hold';
	    }
	    print qq~">\n~;
	}
	print qq~<IMG WIDTH="$list->{$entry}{width}" HEIGHT="$list->{$entry}{height}" SRC="$img" ALT="[$list->{$entry}{alt}]" BORDER="0">~;
	print "</A>" if ($cfg->{options} & ICONS_ARE_LINKS);
	print "</TD>\n";
#Name data
	$msg=$args->{"${mode}_dir"};
	$msg=~s:/$cfg->{rootdir}/:/: if $mode eq 'root' && !$isdir;
	print qq~<TD><A HREF="$msg$entry~;
	print '/' if $isdir;
	if($mode eq 'root') {
	    if($hold) {
		print "?hold_dir=$args->{hold_dir}&hold_file=$args->{hold_file}";
		print "&root_child=" if $stub;
	    } else {
		print "?root_child=" if $stub;
	    }
	    print $stub if $stub;
	    print '#root';
	} elsif($mode eq 'hold') {
	    print "?root_dir=$args->{root_dir}&root_file=$args->{root_file}";
	    print "&hold_child=$stub" if $stub;
	    print '#hold';
	}
	if($entry eq $args->{"${mode}_file"}) {
	    print qq~"><B>$label</B></A></TD>\n~;
	} else {
	    print qq~">$label</A></TD>\n~;
	}
#Last Modified data
	print qq~<TD>$list->{$entry}{modnice}</TD>~ unless ( $cfg->{options} & SUPPRESS_LAST_MOD );
#Size data
	print qq~<TD ALIGN="CENTER">~, $list->{$entry}{sizenice}, "</TD>\n" unless ( $cfg->{options} & SUPPRESS_SIZE );
#Description data
	print '<TD>', $list->{$entry}{desc}, '</TD>' unless ( $cfg->{options} & SUPPRESS_DESC );
	print "</TR>\n";	  
    }
    print "</TABLE><HR>\n";
    if($debug && $dodump) {
	use Data::Dumper;
	print "<PRE>";
	print "\%list\n";
	print Dumper \%$list;
	print "</PRE><HR>";
    }
    return OK;
}

sub thumb_conf {
    my($r) = @_;
    use Storable;
    $config->{cache_dir} = $r->dir_config("IndexCacheDir") || ".thumbnails";
    $config->{dir_create} = $r->dir_config("IndexCreateDir") || 1;
    my $cachedir = $r->filename .  $config->{cache_dir} ;          
    stat $cachedir;
    $config->{cache_ok} = (-e _ && ( -r _ && -w _)) || ((not -e _) && $config->{dir_create} && mkdir $cachedir,0755);
    my $oldopts;
    if ($config->{cache_ok} && -e "$cachedir/.config" && -r _){
	$oldopts = retrieve ("$cachedir/.config");
    }
    $config->{thumb_max_width} = $r->dir_config("ThumbMaxWidth") || DEFAULT_ICON_WIDTH*4;
    $config->{thumb_max_height} = $r->dir_config("ThumbMaxHeight")|| DEFAULT_ICON_HEIGHT*4;
    $config->{thumb_max_size} = $r->dir_config("ThumbMaxSize") || 500000;
    $config->{thumb_min_size} = $r->dir_config("ThumbMinSize") || 5000;
    $config->{thumb_width} = $r->dir_config("ThumbWidth");
    $config->{thumb_height} = $r->dir_config("ThumbHeight");
    $config->{thumb_height} = $r->dir_config("ThumbHeight");
    $config->{thumb_scale_width} = $r->dir_config("ThumbScaleWidth");
    $config->{thumb_scale_height} = $r->dir_config("ThumbScaleHeight");
    $config->{changed} = 0;
    foreach (keys %$config){
	next unless /^thumb/;
	if ($config->{$_} != $oldopts->{$_}) {
	    $config->{changed} = 1;
	    last;
	}
    }
    unless ($config->{cache_ok} && ((not -e "$cachedir/.config") || -w _) && store $config, "$cachedir/.config") {
	$config->{changed} = 0;
    }
}           

sub read_dir {
    my ($r, $dirhandle) = @_;
    my $cfg = Apache::ModuleConfig->get($r);
    my @listing;
    my %list;
    my @accept;
    my $ignore_regex = join('$|^',@{$cfg->{ignore}});
    if($cfg->{options} & THUMBNAILS) {
        #Decode the content-encoding accept field of the client
        foreach (split(',\s*',$r->header_in('Accept'))) {
           push @accept, $_ if m:^image/:o;
    	}
    }
    while(my $file = readdir $dirhandle) {
	next if $file=~m/^\.$|^$ignore_regex$/o; # Never display the '.' directory
	push @listing, $file;
	my $subr = $r->lookup_file($file);
	$list{$file}{uid}=(stat $subr->finfo)[4];
	$list{$file}{gid}=(stat _)[5];
	$list{$file}{size} = -s _;
	if (-d _) {
	    $list{$file}{size} = -1;
	    $list{$file}{sizenice} = '-';
	} else {
	    $list{$file}{sizenice} = size_string($list{$file}{size});
            $list{$file}{sizenice} =~ s/\s*//;    
        }
        $list{$file}{mod}  = (stat _)[9];
        $list{$file}{modnice} = ht_time($list{$file}{mod}, "%d-%b-%Y %H:%M", 0);
        $list{$file}{modnice} =~ s/\s/&nbsp;/g;
        $list{$file}{mode} = write_mod((stat _)[2]);
        $list{$file}{type}  = $subr->content_type;
        if(($list{$file}{type} =~ m:^image/:o) && 
	   ($cfg->{options} & THUMBNAILS ) && 
	   Apache->module("Image::Magick")) {
            if ($config->{cache_ok}) {
                ($list{$file}{icon},$list{$file}{width},$list{$file}{height}) = get_thumbnail($r, $file, $list{$file}{mod}, $list{$file}{type}, @accept);
	    }
	}
        $list{$file}{height} ||= $cfg->{icon_height};
        $list{$file}{width} ||= $cfg->{icon_width};
    #icons size might be calculated on the fly and cached...
	my $icon = Apache::Icon->new($subr);
	$list{$file}{icon} ||= $icon->find;           
	if (-d _) {	
	    $list{$file}{icon} ||= $icon->default('^^DIRECTORY^^');	
	    $list{$file}{alt} = "DIR";
	}	    
	$list{$file}{icon} ||= $icon->default;
        $list{$file}{alt} ||= $icon->alt; 
	$list{$file}{alt} ||= "???"; 
        foreach (keys %{$cfg->{desc}}) {
            $list{$file}{desc} = $cfg->{desc}{$_} if $subr->filename =~ /$_/;
	}
        if($list{$file}{type} eq "text/html" and 
	  ($cfg->{options} & SCAN_HTML_TITLES) and 
	  not $list{$file}{desc}) {
            use HTML::HeadParser;
            my $parser = HTML::HeadParser->new;
            open FILE, $subr->filename;
            while (<FILE>) {
                last unless $parser->parse($_);
	    }
            $list{$file}{desc} = $parser->header('Title');
            close FILE;
	}
        $list{$file}{desc} ||= "&nbsp;";
    }
    return \%list;
}

sub handler {
    my($r)=shift;
    my %args;
    my @items;
    my $name = $r->filename . $r->path_info;
    my $file;
    my $itype;
    my $retval;
    my $fmpath;
    my $uri = $r->uri;
    my $subr;
    my $tail;
    my $cfg = Apache::ModuleConfig->get($r);
    my $postmax=$cfg->{postmax}|DEFAULT_POST_MAX;
    $r = Apache::Request->new($r, POST_MAX=>$postmax);
    $cfg->{rootdir}=DEFAULT_ROOT_DIR unless $cfg->{rootdir};
    my $rootdir=$cfg->{rootdir};
    $cfg->{holddir}=DEFAULT_HOLD_DIR unless $cfg->{holddir};
    my $holddir=$cfg->{holddir};
    ($fmroot)=$uri=~m:(^.*/)$rootdir/:;
    $fmroot=$cfg->{root} unless $fmroot;
    $debug=$cfg->{debug} || $r->dir_config('AutoIndexDebug'); 
    print STDERR "===== ", __PACKAGE__, " DEBUG START =====\nuri=$uri " if $debug;
    $name .= '/' unless $name =~ m:/$:o;
    if($name=~m:/$rootdir/:) {
	$hold=0;
	($fmpath)=$name=~m:(^.*)/$rootdir/:;
	unless(-d "$fmpath/$rootdir") {
	    $r->log_reason( __PACKAGE__ . " Path not found:");
	    print STDERR "FORBIDDEN\n===== ", __PACKAGE__, " DEBUG STOP  =====\n" if $debug;
	    return FORBIDDEN;
	}
	if($name=~m:/$rootdir/$holddir/:) {
	    ($tail)=$name=~m:$rootdir/$holddir/(.*/?)$:;
	    $itype.='h';
	} else {
	    ($tail)=$name=~m:$rootdir/(.*/?)$:;
	    $itype.='x';
	    if($cfg->{hold} && -d "$fmpath/$rootdir/$holddir/") {
		$hold=1;	# if hold directory and hold directive
	    }
	    $name="$fmpath/$tail";
	}
	print STDERR "name=$name type=$itype root=$rootdir hold=$holddir\n" if $debug;
    }
    $name=~s:/$::;							# Remove any trailing '/'
    $subr = $r->lookup_file("$name");
    stat $subr->finfo;
    unless(-e _) {
	$r->log_reason( __PACKAGE__ . " Path not found: ");
	print STDERR "FORBIDDEN\n===== ", __PACKAGE__, " DEBUG STOP  =====\n" if $debug;
	return FORBIDDEN;
    } elsif(-d _) {
	unless ($r->path_info || $tail) { #Issue an external redirect if the dir isn't tailed with a '/'
	    $r->header_out(Location=>"$uri/");
	    $nRedir++;
	    print STDERR "REDIRECT\n===== ", __PACKAGE__, " DEBUG STOP  =====\n" if $debug;
	    return REDIRECT;
	}
	$name .= '/' unless $name =~ m:/$:o;
	$itype.='d' if $itype;
    } else {
	($file)=$name =~ m:.*/(.+)$:o;	# filename clicked
	if($itype) {
	    $itype.='f';			# not a directory, but a file
	    $name =~ s:(.+/).*:$1:;	# the directory clicked
	}
    }
    $r->filename("$name");
    unless($fmpath || ($r->content_type && $r->content_type eq DIR_MAGIC_TYPE)) {
	print STDERR "DECLINED\n===== ", __PACKAGE__, " DEBUG STOP  =====\n" if $debug;
	return DECLINED;
    }
	@$gid=();	# Remove all groups
    if($r->allow_options & OPT_INDEXES || $itype) {
	if($itype) {
	    $args{form}='';
	    my @params = $r->param;
	    foreach my $arg (@params) {
		my @values=$r->param($arg);
		foreach my $value (@values) {
		    if($value eq 'on') {
			push @items, $arg;
		    } else {
			$args{$arg}=$value;
		    }
		}
	    }
	    if($r->auth_type eq 'Basic') {
		$user=$r->user;
	    } elsif($cfg->{userenv}) {
		$user=$r->subprocess_env($cfg->{userenv});
	    } else {
		$user='';
	    }
	    my $gidenv = $cfg->{gidenv};
	    if($gidenv) {
		@$gid=split /[,:;]|$;/,$r->subprocess_env($gidenv);
		if($debug) {
		    print STDERR "GID: gidenv=$gidenv gids=";
		    for(my $cnt=@$gid-1;$cnt>=0;$cnt--) {
				print STDERR "$gid->[$cnt],";
		    }
		    print STDERR "\n";
		}
		for(my $cnt=@$gid-1;$cnt>=0; $cnt--) {
		    if($gid->[$cnt]=~m:[^0-9]:o) {		# if not a number look-up the group name
			$gidname->[$cnt]=$gid->[$cnt];
			$gid->[$cnt]=getgrnam $gidname->[$cnt];
		    } else {
			$gidname->[$cnt]=getgrgid $gid->[$cnt] || $gid->[$cnt];
		    }
		}
	    }
	    if(@$gid) {
		$admin  = $cfg->{admin};
		$umask = $cfg->{umask};
		print STDERR "GID: admin=$admin umask=$umask\n" if $debug;
	    }
	    my $docroot=$r->document_root;
	    if($cfg->{revoke} && @$gid) {
		my $server=$r->get_server_name;
		unless($Users->{"##${server}#${fmroot}"} eq '~') {	# Initialize Tag
		    getrevoked($r,"$docroot$fmroot$rootdir".REVOKE_DIR.REVOKE_FILE);
		    $Users->{"##${server}#${fmroot}"}='~';
		}
		if($Users->{"$user##$server#$fmroot"} eq '-') {
		    $gid=(-1);	# Remove all groups
		} else {
		    for(my $cnt=@$gid-1;$cnt>=0;$cnt--) {
			$gid->[$cnt]=-1 if $Users->{"#$gid->[$cnt]#$server#$fmroot"} eq '-';
		    }
		}
	    }
	    if($itype=~m:h:o) {
		$args{hold_child}="" if $itype=~m:f:o;
		$args{hold_file}=$file;
		($args{hold_dir})=$uri=~m:(.*/).*$:o;	# strip any filename
	    }
	    if($itype=~m:x:o) {
		$args{root_child}="" if $itype=~m:f:o;
		$args{root_file}=$file;
		($args{root_dir})=$uri=~m:(.*/).*$:o;	# strip any filename
	    }
		$dodump=$debug unless $itype;	# Turn on dump for AutoIndex mode
	    if($args{proc}) {
		if($args{proc} eq 'revoke') {
		    my  $revgid=$args{id} if $args{enagid} || $args{disgid};
		    my $revuser=$args{id} if $args{enauid} || $args{disuid};
		    if($revgid=~m:[A-Za-z]:) {
			$revgid=getgrnam $revgid;
		    }
		    if($revuser eq $user || $revgid==$admin) {
			$r->warn(__PACKAGE__ . " revoke self not allowed");
			errmsg("admin IDs can not be revoked");
			$args{root_err}=1;
		    } else {
			my $result;
			$file="$docroot$fmroot$rootdir".REVOKE_DIR.REVOKE_FILE;
			$result=revoke($r,$cfg,'user','+',$args{id},$file) if $args{enauid};
			$result=revoke($r,$cfg,'user','-',$args{id},$file) if $args{disuid};
			$result=revoke($r,$cfg,'gid', '+',$args{id},$file) if $args{enagid};
			$result=revoke($r,$cfg,'gid', '-',$args{id},$file) if $args{disgid};
			$result=1 if $args{return};
			unless($result) {
			    $r->warn(__PACKAGE__ . " Revoke: $user: $args{id}: $file");
			    $args{root_err}=1;
			} else {
			    $r->log->notice(__PACKAGE__." Revoke: $user: $args{id}: $file");
			}
		    }
		    delete $args{id};
		}
		if($args{proc} eq 'edit') {
		    if($args{save}) {
			if(EditSave($r,$cfg,$docroot,\%args)) {
			    $r->log->notice(__PACKAGE__." $user: EditSave: $args{edit_file}");
			}
		    }
		    my    $file="$docroot$args{file}";
		    my $inifile="$docroot$args{info}";
		    if($args{save} || $user eq $args{user}) {
			if($args{save}) {
			    unless(open INIFILE, ">$inifile") {
				errmsg("Edit: Lock File write open");
				$args{hold_err}=1 if $args{form} eq 'hold';
				$args{root_err}=1 if $args{form} eq 'root';
			    } else {
				unless(flock INIFILE, LOCK_EX|LOCK_NB) {
				    errmsg("Edit: Couldn't lock file.  Try again");
				    $args{hold_err}=1 if $args{form} eq 'hold';
				    $args{root_err}=1 if $args{form} eq 'root';
				} else {
				    my $fgid=(stat $file)[5];
	 			    $fgid=getgrgid $fgid || $fgid;
				    print INIFILE "editedby=$user\ngid=$fgid\ntime=",scalar localtime,"\nstatus=in\n";
				}
			    }
			} else {
			    unless(open INIFILE, ">>$inifile") {
				errmsg("Edit: Lock File append open");
				$retval=HOLD_ERR if $args{form} eq 'hold';
				$retval=ROOT_ERR if $args{form} eq 'root';
			    } else {
				unless(flock INIFILE, LOCK_EX|LOCK_NB) {
				    errmsg("Edit: Couldn't lock file.  Try again");
				    $args{hold_err}=1 if $args{form} eq 'hold';
				    $args{root_err}=1 if $args{form} eq 'root';
				} else {
				    print INIFILE "status=in\n";
				}
			    }
			}
			flock INIFILE,LOCK_UN;
			close INIFILE;
			delete $args{text};
			delete $args{file};
			delete $args{info};
			delete $args{user};
		    }
		}
		if($args{proc} eq 'hold' || $args{proc} eq 'root') {
		    unless($args{hold_dir}) {	# Fixup hold_dir if missing
			$args{hold_dir}="$fmroot$rootdir/$holddir/";
		    }
		    unless($args{root_dir}) {	# Fixup root_dir if missing
			$args{root_dir}="$fmroot$rootdir/";
		    }
		    $retval=procform($r,\%args,\@items) unless $retval;
		    if($retval>99) {
			$nRedir++;
			print STDERR "REDIRECT $retval\n===== ", __PACKAGE__, " DEBUG STOP  =====\n" if $debug;
			return  $retval;
		    } 
		    $args{hold_err}=1 if $retval==HOLD_ERR;
		    $args{root_err}=1 if $retval==ROOT_ERR;
		}
	    }
	}
	unless($retval==SKIP_INDEX) {
	    $retval=file_man($r,\%args,$itype);
	} else {
	    $retval=OK;
	}
	if($debug && $dodump) {
	    use Data::Dumper;
	    print "<PRE>\n\$dodump=$dodump\n";
	    print "\$cfg\n";
	    print Dumper $cfg;
	    print "</PRE><HR><PRE>\%args\n";
	    print Dumper \%args;
    	    if(@items) {
		print "</PRE><HR><PRE>\@items\n";
		print Dumper \@items;
	    }
	    print "</PRE><HR><PRE>Environment variables\n";
	    print Dumper $r->subprocess_env();
	    print "</PRE><HR><PRE>Global variables\n";
	    print "\$user=$user\n" if $user;
	    if(@$gid) {
		print "\$gid\n";
		print Dumper $gid;
		print "\$gidname\n";
		print Dumper $gidname;
	    }
	    print "\$admin=$admin\n" if $admin;
	    print "\$umask=$umask\n" if $umask;
	    print "\$fmroot=$fmroot\n\$hold=$hold\n\$errmsg=$errmsg\n";
	    if($cfg->{revoke}) {
		print "\$Users\n";
		print Dumper $Users;
	    }
	    print "\$config\n";
	    print Dumper $config;
	    print "</PRE><HR>\n";
	}
	footer($r);
	print STDERR "retval=$retval\n===== ", __PACKAGE__, " DEBUG STOP  =====\n" if $debug;
	return $retval;
    } else {
	$r->log_reason( __PACKAGE__ . " Directory index forbidden by rule", $r->uri . " (" . $r->filename . ")");
	print STDERR "FORBIDDEN\n===== ", __PACKAGE__, " DEBUG STOP  =====\n" if $debug;
	return FORBIDDEN;
    }
}

sub getrevoked {
    my ($r,$file)=@_;
    my $server=$r->get_server_name;
    my $key;
    print STDERR "getrevoked() file=$file\n" if $debug;
    if(open REVOKED, $file) {
	while(<REVOKED>) {
	    my($type,$val)=m:(\w+)=(\w+):;
	    $val = lc $val;
		 $key=$type eq 'user'?"$val#":'#';
	    $key.=$type eq 'gid' ?"$val#":'#';
		$key.="${server}#${fmroot}";
	    print STDERR "getrevoked() found $type=$val key=$key\n" if $debug;
		$Users->{"$key"}='-';
	}
	close REVOKED;
    }
}

sub do_sort {
    my ($list, $query, $default) = @_;
    my @names = sort keys %$list;
    shift @names;                   #removes '..'
#handle default sorting
    unless ($query->{N} || $query->{S} || $query->{D} || $query->{M}) {
	$default =~ /(.)(.)/;
	$query->{$1} = $2;
    }
    if ($query->{N}) {
	@names = sort file_ext @names if $query->{N} eq 'E';
	@names = sort @names if $query->{N} eq 'D';
	@names = reverse sort @names if $query->{N} eq "A";
    } elsif ($query->{S}) {
	@names = sort { $list->{$b}{size} <=> $list->{$a}{size} } @names if $query->{S} eq "D";
	@names = sort { $list->{$a}{size} <=> $list->{$b}{size} } @names if $query->{S} eq "A";
    } elsif ($query->{M}) {
	@names = sort { $list->{$b}{mod} <=> $list->{$a}{mod} } @names if $query->{M} eq "D";
	@names = sort { $list->{$a}{mod} <=> $list->{$b}{mod} } @names if $query->{M} eq "A";		
    } elsif ($query->{D}) {
	@names = sort { $list->{$b}{desc} cmp $list->{$a}{desc} } @names if $query->{D} eq "D";
	@names = sort { $list->{$a}{desc} cmp $list->{$b}{desc} } @names if $query->{D} eq "A";		
    }
    unshift @names, '..';           #puts back '..' on top of the pile
    return \@names;
}

sub file_ext {
	my @as=split /\./,$a;
	my @bs=split /\./,$b;
    my $retval=0;
    my $cnt=$#as;
	$cnt=$#bs if @bs>@as;
    for(;$cnt>=0;$cnt--) {
	$retval=$bs[$cnt] cmp $as[$cnt]; 
	return $retval if $retval;
    }
    0;
}

sub get_thumbnail {
    my ($r, $filename, $mod, $content, @accept) = @_; 
    my $accept = join('|', @accept);
    my $dir = $r->filename;
#these should sound better.
    my $cachedir = $config->{cache_dir};
    my $xresize;
    my $yresize;
    my $img = Image::Magick->new;
    my($imgx, $imgy, $img_size, $img_type) = split(',', $img->Ping($dir . $filename));
#Is the image OK?
    return "/icons/broken.gif" unless ($imgx && $imgy);
    if (($content =~ /$content/) && ($img_type =~ /JPE?G|GIF|PNG/i)) {
	if ($dir =~ /$cachedir\/$/) {	#We know that what we'll generate will be seen.
	    return $filename, $imgx, $imgy #Avoiding recursive thumbnails from Hell
	}
	return undef if $img_size > $config->{thumb_max_size}; #The image is way too big to try to process...
	if(defined $config->{thumb_scale_width} || 
           defined $config->{thumb_scale_height}) {
            #Factor scaling
            $xresize = $config->{thumb_scale_width} * $imgx if defined $config->{thumb_scale_width};
            $yresize = $config->{thumb_scale_height} * $imgy if defined $config->{thumb_scale_height};           
	} elsif(defined $config->{thumb_width} || 
	    defined $config->{thumb_height}) {
#Absolute scaling
	    $xresize = $config->{thumb_width}  
		if defined $config->{thumb_width};
	    $yresize = $config->{thumb_height} 
		if defined $config->{thumb_height};           
	}
#preserve ratio if we can
	$xresize ||= $yresize * ($imgx/$imgy);
	$yresize ||= $xresize * ($imgy/$imgx);   
#default if values are missing.
	$xresize ||= DEFAULT_ICON_WIDTH;
	$yresize ||= DEFAULT_ICON_HEIGHT;
#round off for picky browsers
	$xresize = int($xresize);
	$yresize = int($yresize);
#Image is too small to actually resize.  Simply resize with the WIDTH and HEIGHT attributes of the IMG tag
	return ($filename, $xresize , $yresize) if $img_size < $config->{thumb_min_size};
	if ($config->{changed} || $mod > (stat "$dir$cachedir/$filename")[9]) {
#We should actually resize the image
	if ($img->Read($dir . $filename)) { #Image is broken
	    return "/icons/broken.gif";
	}
	$nThumb++;
	$img->Sample(width=>$xresize, height=>$yresize);
	$img->Write("$dir$cachedir/$filename");       
    }
    return "$cachedir/$filename", $xresize , $yresize;
    }   
    return undef;
}

sub place_doc {
    my ($r,$cfg,$docs) = @_;
    my $uri = $r->uri;
    $uri=~s:/$cfg->{rootdir}/:/: if $cfg->{root_dir};
    foreach my $doc (@{$cfg->{$docs}}) {
	print STDERR "place_doc() $docs=" if $debug;
	my $subr = $r->lookup_uri("${uri}${doc}");
	if (stat $subr->finfo) {
	    print STDERR "${uri}${doc}\n" if $debug;
	    print "<PRE>" unless $doc=~m/\.html/;
	    $subr->run;
	    print "</PRE>" unless $doc=~m/\.html/;
	    print "<HR>";
	    next;
	}
	$subr = $r->lookup_uri("${uri}${doc}.html");
	if (stat $subr->finfo) {
	    print STDERR "${uri}${doc}.html\n" if $debug;
	    $subr->run;
	    print "<HR>";
	    next;
	}
	$subr = $r->lookup_uri("/$doc");
	if (stat $subr->finfo) {
	    print STDERR "${doc}\n" if $debug;
	    print "<PRE>" unless $doc=~m/\.html/;
	    $subr->run;
	    print "</PRE>" unless $doc=~m/\.html/;
	    print "<HR>";
	    next;
	}
	$subr = $r->lookup_uri("/${doc}.html");
	if (stat $subr->finfo) {
	    print STDERR "${doc}.html\n" if $debug;
	    $subr->run;
	    print "<HR>";
	    next;
	}
	print STDERR "<MISSING>\n" if $debug;
    }
}

sub write_mod {
    my $mod = shift ;
    $mod = $mod & 4095;
    my $letters;
    my %modes = (
	1   =>  'x',
	2   =>  'w',
	4   =>  'r',
    );
    foreach my $f (64,8,1){
        foreach my $key (4,2,1) {
	    if ($mod & ($key * $f)){
                $letters .= $modes{$key};
	    } else {
		$letters .= '-';
	    }
	}
    }
    return $letters;
}

#Configuration Stuff
sub patternize{
    my $pattern = shift;
    $pattern =~ s/\./\\./g;
    $pattern =~ s/\*/.*/g;
    $pattern =~ s/\?/./g;
    return $pattern;
}

sub push_config{
    my ($cfg, $parms, $value) = @_;
    my $key = $parms->info;
    if ($key eq 'ignore'){
	$value = patternize($value);
    }
    push @ {$cfg->{$key}}, $value;
    return DECLINE_CMD if Apache->module('mod_autoindex.c');
}

sub DirectoryIndex($$$;*){
    my ($cfg, $parms, $files, $cfg_fh) = @_;
    for my $file (split /\s+/, $files) {
	push @{$cfg->{indexfile}}, $file;
    }
    return DECLINE_CMD if Apache->module('mod_dir.c');
}

sub IndexOptions($$$;*){
    my ($cfg, $parms, $directives, $cfg_fh) = @_;
    foreach (split /[\s,]+/, $directives) {
	my $option;
	(my $action, $_) = (lc $_) =~ /(\+|-)?(.*)/;
	if (/^none$/) {
	    die "Cannot combine '+' or '-' with 'None' keyword" if $action;
	    $cfg->{options} = NO_OPTIONS;
	    $cfg->{options_add} = 0;
	    $cfg->{options_del} = 0;
	} elsif (/^iconheight(=(\d*$|\*$)?)?(.*)$/) {
	    warn "Bad IndexOption $_ directive syntax" if ($3 || ($1 && !$2));
	    if ($2) {
		die "Cannot combine '+' or '-' with IconHeight" if $action;
		$cfg->{icon_height} = $2;
	    } else 	{
		if ($action eq '-') {
		    $cfg->{icon_height} = DEFAULT_ICON_HEIGHT;
		} else {
		    $cfg->{icon_height} = 0;
		}
	    }
	} elsif (/^iconwidth(=(\d*$|\*$)?)?(.*)$/) {
	    warn "Bad IndexOption $_ directive syntax" if ($3 || ($1 && !$2));
	    if ($2) {
		die "Cannot combine '+' or '-' with IconWidth" if $action;
		$cfg->{icon_width} = $2;
	    } else {
		if ($action eq '-') {
		    $cfg->{icon_width} = DEFAULT_ICON_WIDTH;
		} else {
		    $cfg->{icon_width} = 0;
		}
	    }
	} elsif (/^namewidth(=(\d*$|\*$)?)?(.*)$/) {
	    warn "Bad IndexOption $_ directive syntax" if ($3 || ($1 && !$2));
	    if ($2) {
		die "Cannot combine '+' or '-' with NameWidth" if $action;
		$cfg->{name_width} = $2;
	    } else {
		die "NameWidth with no value can't be used with '+'" if ($action ne '-');
		$cfg->{name_width} = 0;
	    }
	} else {
	    foreach my $directive (keys %GenericDirectives) {
		if(/^$directive$/) {
		    $option = $GenericDirectives{$directive};
		    last;                
		}
	    }
	    warn "IndexOptions unknown/unsupported directive $_" unless $option;
	}
	if (! $action) {
	    $cfg->{options} |= $option;
	    $cfg->{options_add} = 0;
	    $cfg->{options_del} = 0;
	} elsif ($action eq '+') {
	    $cfg->{options_add} |= $option;
	    $cfg->{options_del} &= ~$option;
	} elsif ($action eq '-') {
	    $cfg->{options_del} |= $option;
	    $cfg->{options_add} &= ~$option;
	}
	if (($cfg->{options} & NO_OPTIONS) && ($cfg->{options} & ~NO_OPTIONS)) {
	    die "Cannot combine other IndexOptions keywords with 'None'";
	}
    }
    return DECLINE_CMD if Apache->module('mod_autoindex.c');
}

sub rmarray {
    my ($array, $element) = @_;
    for(my $ndx; $ndx<@$array; $ndx++) {
	return splice @$array,$ndx,1 if lc @$array[$ndx] eq lc $element;
    }
}

sub FileManOptions($$$;*) {
    my ($cfg, $parms, $directive, $cfg_fh) = @_;
    my @args=split /[\s=,]+/, $directive;
    unless($args[0]) {
	warn "FileManOptions $directive directive: No argument";
    }
    my $lcarg = lc shift @args;
    my ($action)=$args[0]=~m:^([+-]):o;
    $args[0]=~s:^[+-]::o if $action;
    my $arg=$args[0];
    if($lcarg eq 'rootmenu') {
	splice @{$cfg->{rootmenu}} unless $action; # clear the array
        foreach (@args) {
	    if($action eq '-') {
		rmarray \@{$cfg->{rootmenu}},$_;
	    } else {
		unshift @{$cfg->{rootmenu}},$_;
	    }
	}
    } elsif($lcarg eq 'holdmenu') {
	splice @{$cfg->{holdmenu}} unless $action;
	foreach (@args) {
	    if($action eq '-') {
		rmarray \@{$cfg->{holdmenu}},$_;
	    } else {
		unshift @{$cfg->{holdmenu}},$_;
	    }
	}
    } elsif($lcarg eq 'admnmenu') {
	splice @{$cfg->{admnmenu}} unless $action;
	foreach (@args) {
	    if($action eq '-') {
		rmarray \@{$cfg->{admnmenu}},$_;
	    } else {
		unshift @{$cfg->{admnmenu}},$_;
	    }
	}
    } elsif ($lcarg eq 'textlen') {
	if($arg<8) {
	    warn "Bad FileManOptions $directive directive<8";
	    $cfg->{textlen} = DEFAULT_TEXT_LEN;
	} else {
	    $cfg->{textlen} = $arg;
	}
    } elsif ($lcarg eq 'postmax') {
	if($arg<128000) {
	    warn "Bad FileManOptions $directive directive<128000";
	    $cfg->{postmax} = DEFAULT_POST_MAX;
	} else {
	    $cfg->{postmax} = $arg;
	}
    } elsif ($lcarg eq 'editmax') {
	if($arg<1024) {
	    warn "Bad FileManOptions $directive directive<1024";
	    $cfg->{postmax} = DEFAULT_EDIT_MAX;
	} else {
	    $cfg->{postmax} = $arg;
	}
    } elsif ($lcarg eq 'admin') {
	if($arg=~m:[^0-9]:o) {
	    $arg=getgrnam $arg;
	}
	$cfg->{admin}=$arg;
    } elsif ($lcarg eq 'umask') {
	if($arg>0777 || $arg<0001) {
	    warn "Bad FileManOptions $directive directive";
	} else {
	    $cfg->{umask} = $arg;
	}
    } elsif ($lcarg eq 'help') {
	$cfg->{help} = $arg;
    } elsif ($lcarg eq 'debug') {
	$arg = lc $arg;
	if($arg eq '1' || $arg eq 'yes' || $arg eq 'on') {
	    $cfg->{debug} = 1;
	} else {
	    $cfg->{debug} = 0;
	}
    } elsif ($lcarg eq 'language') {
	$arg = lc $arg;
	if($arg eq '1' || $arg eq 'yes' || $arg eq 'on') {
	    $cfg->{language} = 1;
	} else {
	    $cfg->{language} = 0;
	}
    } elsif ($lcarg eq 'hold') {		# Force hold directory
	$arg = lc $arg;
	if($arg eq '1' || $arg eq 'yes' || $arg eq 'on') {
	    $cfg->{hold} = 1;
	} else {
	    $cfg->{hold} = 0;
	}
    } elsif ($lcarg eq 'revoke') {
	$arg = lc $arg;
	if($arg eq '1' || $arg eq 'yes' || $arg eq 'on') {
	    $cfg->{revoke} = 1;
	} else {
	    $cfg->{revoke} = 0;
	}
    } elsif ($lcarg eq 'nocache') {
	$arg = lc $arg;
	if($arg eq '1' || $arg eq 'yes' || $arg eq 'on') {
	    $cfg->{nocache} = 1;
	} else {
	    $cfg->{nocache} = 0;
	}
    } elsif ($lcarg eq 'userenv') {
	$cfg->{userenv} = $arg;
    } elsif ($lcarg eq 'gidenv') {
	$cfg->{gidenv} = $arg;
    } elsif ($lcarg eq 'root') {
	$cfg->{root} = $arg;
    } else {
	$arg=~s:/$::o; # Remove any trailing '/'
	if($lcarg eq 'rootdir') {
	    $cfg->{rootdir}=$arg;
	} elsif ($lcarg eq 'holddir') {
	    $cfg->{holddir} = $arg;
	} else {
	    warn "Unknown FileManOptions $directive directive";
	}
    }
}

# e.g. DirectoryIndex index.html index.htm index.cgi 

sub AddDescription($$$;*){
#this is not completely supported.  
#Since I didn't take the time to fully check mod_autoindex.c behavior,
#I just implemented this as simplt as I could.
    my ($cfg, $parms, $args, $cfg_fh) = @_;
    my ($desc, $files) = ( $args =~ /^\s*"([^"]*)"\s+(.*)$/);
    my $file = join "|", split /\s+/, $files;
    $file = patternize($file);
    $cfg->{desc}{$file} = $desc; 
    return DECLINE_CMD if Apache->module('mod_autoindex.c');
}

sub IndexOrderDefault($$$$){
    my ($cfg, $parms, $order, $key) = @_;
    die "First Keyword must be Ascending or ending" unless ( $order =~ /^(de|a)scending$/i);
    die "First Keyword must be Name, Date, Size or Description" unless ( $key =~ /^(date|name|size|description)$/i);
    if ($key =~ /date/i) {
	$key = 'M';
    } else {
	$key =~ s/(.).*$/$1/;
    }
    $order =~ s/(.).*$/$1/;
    $cfg->{default_order} = $key . $order;
    return DECLINE_CMD if Apache->module('mod_autoindex.c');
}

sub FancyIndexing ($$$) {
    my ($cfg, $parms, $opt) = @_;
    die "FancyIndexing directive conflicts with existing IndexOptions None" if ($cfg->{options} & NO_OPTIONS);
    $cfg->{options} = ( $opt ? ( $cfg->{options} | FANCY_INDEXING ) : ($cfg->{options} & ~FANCY_INDEXING ));
    return DECLINE_CMD if Apache->module('mod_autoindex.c');
}
	
sub DIR_CREATE {
    my $class = shift;
    my $self = $class->new;
    $self->{icon_width} = DEFAULT_ICON_WIDTH;
    $self->{icon_height} = DEFAULT_ICON_HEIGHT;
    $self->{name_width} = DEFAULT_NAME_WIDTH;
    $self->{default_order} = DEFAULT_ORDER;
    $self->{rootdir} = DEFAULT_ROOT_DIR;
    $self->{holddir} = DEFAULT_HOLD_DIR;
    $self->{root}  = "";
    $self->{admin} = 0;
    $self->{umask} = 0;
    $self->{rootmenu} = DEFAULT_ROOT_MENU;
    $self->{holdmenu} = DEFAULT_HOLD_MENU;
    $self->{admnmenu} = DEFAULT_ADMN_MENU;
    $self->{hold} = 0;
    $self->{revoke}  = 0;
    $self->{textlen} = DEFAULT_TEXT_LEN;
    $self->{postmax} = DEFAULT_POST_MAX;
    $self->{help} = DEFAULT_HELP_URL;
    $self->{language}=0;
    $self->{gidenv} = "";
    $self->{userenv}= "";
    $self->{ignore} = [];
    $self->{readme} = [];
    $self->{header} = [];
    $self->{indexfile} = [];
    $self->{desc} = {};
    $self->{options} = 0;
    $self->{options_add} = 0;
    $self->{options_del} = 0;
    return $self;
}

sub new { 
    return bless {}, shift;
}

sub DIR_MERGE {
    my ($parent, $current) = @_;
    my %new;
    $new{options_add}   = 0;
    $new{options_del}   = 0;
    $new{icon_height}   = $current->{icon_height}   || $parent->{icon_height};
    $new{icon_width}    = $current->{icon_width}    || $parent->{icon_width};
    $new{name_width}    = $current->{name_width}    || $parent->{name_width};
    $new{default_order} = $current->{default_order} || $parent->{default_order};
    $new{rootdir}  = $current->{rootdir}  || $parent->{rootdir};
    $new{holddir}  = $current->{holddir}  || $parent->{holddir};
    $new{root}     = $current->{root}     || $parent->{root};
    $new{admin}    = $current->{admin}    || $parent->{admin};
    $new{umask}    = $current->{umask}    || $parent->{umask};
    $new{textlen}  = $current->{textlen}  || $parent->{textlen};
    $new{postmax}  = $current->{postmax}  || $parent->{postmax};
    $new{help}     = $current->{help}     || $parent->{help};
    $new{language} = $current->{language} || $parent->{language};
    $new{userenv}  = $current->{userenv}  || $parent->{userenv};
    $new{gidenv}   = $current->{gidenv}   || $parent->{gidenv};
    $new{hold}     = $current->{hold}     || $parent->{hold};
    $new{revoke}   = $current->{revoke}   || $parent->{revoke};
    $new{rootmenu} = $current->{rootmenu} || $parent->{rootmenu};
    $new{holdmenu} = $current->{holdmenu} || $parent->{holdmenu};
    $new{admnmenu} = $current->{admnmenu} || $parent->{admnmenu};
    $new{readme}    = [ @{$current->{readme}},    @{$parent->{readme}} ];
    $new{header}    = [ @{$current->{header}},    @{$parent->{header}} ];
    $new{ignore}    = [ @{$current->{ignore}},    @{$parent->{ignore}} ];
    $new{indexfile} = [ @{$current->{indexfile}}, @{$parent->{indexfile}} ];
    $new{desc} = {% {$current->{desc}}};    #Keep descriptions local
    if ($current->{options} & NO_OPTIONS) { #None override all directives
	$new{options} = NO_OPTIONS;
    } else {
	if ($current->{options} == 0) { #Options are all incremental, so combine them with parent's values
	    $new{options_add} = ( $parent->{options_add} | $current->{options_add}) & ~$current->{options_del};
	    $new{options_del} = ( $parent->{options_del} | $current->{options_del}) ;
	    $new{options} = $parent->{options} & ~NO_OPTIONS;
	} else { #Options weren't all incremental, so forget about inheritance, simply override
	    $new{options} = $current->{options};
	}
        $new{options} |= $new{options_add};
	$new{options} &= ~ $new{options_del};
    }
    return bless \%new, ref($parent);
}

sub status {
    my ($r, $q) = @_;
    my @s;
    my $cfg = Apache::ModuleConfig->get($r);
    push (@s, "<B>" , __PACKAGE__ , " (ver $Apache::FileMan::VERSION) statistics</B><BR>");
    push (@s , "Done " . $nDir . " listings so far<BR>");
    push (@s , "Done " . $nRedir . " redirects so far<BR>");
    push (@s , "Done " . $nIndex. " indexes so far<BR>");
    push (@s , "Done " . $nThumb. " thumbnails so far<BR>");
    use Data::Dumper;
    my $string = Dumper $cfg;
    push (@s, $string);
    return \@s;
}

1;
__END__
=head1 NAME

Apache::FileMan - Perl File Manager for Apache Web server

=head1 SYNOPSIS

  PerlModule Apache::Icon
  PerlModule Apache::FileMan
  (PerlModule Apache::Language) optional
  (PerlModule Image::Magick)    optional

=head1 DESCRIPTION

FileMan provides a file manager for a web sites through a web
browser. It is a extensive rewrite of the Apache::AutoIndex.pm
module which in turn was a remake of the autoindex Apache
module. FileMan can provide the same functionality as
AutoIndex.pm and can be used to both navigate and manage the web
site.

FileMan has dropped the mod_dir support provided by AutoIndex.

I order to activate the file manager functionality, two things
have to happen. First, the proper http.conf directives need to
be placed into a <Location area> section. Second, there has to
be a directory stub (.XFM) created off of the directory where
the file manager is to be provided.

Within the ROOT directory stub (.XFM), a HOLD sub-directory
(.XFM/hold) can also be provided to present a HOLD directory
tree by the file manager. The HOLD (.XFM/hold) directory
provides a physical holding area where files can be uploaded,
unzipped, moved, copied, deleted, and renamed. New directories
can be created with the mkdir command. The HOLD directory is
normally were a user would organize their site files before
updating (copy/move) the site's ROOT directory. In order to
activate the HOLD directory (tree) display the "hold" directive
needs to be set to '1'.  When both the ROOT and HOLD directory
trees are displayed, it is known as the "double tree" mode.

The ROOT (.XFM) directory is actually a fake path of the site's
root directory. For example to access FileMan: "http:
//www.site.com/bob/" the following URL would be required:

	"http://www.site.com/bob/.XFM/"

This would in turn would display the file manager for bob. To
Bob, the ROOT directory appears to be his actual web root
directory.

If the above description does not make sense, just follow the
examples provided, and perhaps it will become clearer once you
see some results.

Since a URL fake path (.XFM) is provided, authentication and
authorization can be used to only allow authorized users to
have access to the FileMan module.

In short, you will no longer need to use ftp to upload and
manage the web site files. Since FileMan is web based, you can
use all of your other Apache functionality, such as SSL,
proxies, and etc.
  
The best procedure to get FileMan loaded and working is to first
have the Apache mod_perl and autoindex modules loaded and
working properly. Then remove the httpd.conf "AddModule
autoindex" directive and add the Apache::Icon and 
Apache::FileMan module directives.
  
=head1 DIRECTIVES

=head2 Loading the Modules

The following describes what httpd.conf directives you need in
your httpd.conf file to load FileMan and it's companion modules.

First or all you must have mod_perl loaded, with the following:

AddModule mod_perl.c

You will also need to load the following mod_perl modules, with:

  PerlModule Apache::Icon
  PerlModule Apache::FileMan

in your httpd.conf file or with:

   use Apache::Icon();
   use Apache::FileMan();
 
in your starup.pl file.

=head2 Configuration Guidelines

It is best to put the FileMan directives is in a <Location area>
section of your httpd.conf file, because it is the highest
priority Apache httpd.conf section. This way, other directives
will not get in the way of (ahead of) FileMan during the Apache
request processing. Apache 1.3.x the directive section priorities
are (in increasing order):

    <Directory>
    <Files>
    <Location>

Here is an example of a <Location area> directive:

    <LocationMatch /.*/\.XFM>
	SetHandler perl-script
	PerlHandler Apache::FileMan
    </LocationMatch> 

Notice that a regular expression Location form was used. This
will provide a file manager for each 1-level deep
sub-directory of the site's document root which have a
.XFM stub directory in them.  For example:
    
http:://www.site.com/friends/bob/

If a browser in turn accesses:

    http:://www.site.com/friends/bob/.XFM/

The FileMan file manager would be activated for "/friends/bob".

Even though the .XFM directory is a fake reference for the real
directory tree, it must exist in order to activate the file
manager. If a ".XFM/hold" directory is also present, and the
"hold" directive is set to '1', the double tree form of FileMan
will be presented.  Normally you will only want to activate the
double tree mode for advanced users.

You will probably want to provide authentication and
authorization for the .XFM fake location. For example, I have
used Apache::AuthenDBI and Apache::AuthzDBI with the following
additions to the same <Location> as above:

 PerlAuthenHandler Apache::AuthenDBI
 PerlAuthzHandler  Apache::AuthzDBI
 AuthName DBI
 AuthType Basic
 PerlSetVar Auth_DBI_data_source  dbi:Pg:dbname=webdb
 PerlSetVar Auth_DBI_username     webuser
 PerlSetVar Auth_DBI_password     webpass
 PerlSetVar Auth_DBI_pwd_table    users
 PerlSetVar Auth_DBI_uid_field    username
 PerlSetVar Auth_DBI_grp_field    GID
 PerlSetVar Auth_DBI_pwd_field    password
 PerlSetVar Auth_DBI_encrypted    on
 require group webgroup friends propellers

If you only want to provide the AutoIndex functionality, just place the
following into either a <Directory area>, or <Location area>
directive and don't bother to create the .XFM directory.

 SetHandler perl-script
 PerlHandler Apache::FileMan

Mod_perl does not provide configuration merging for Apache
virtual hosts. Therefore, you have to maintain a complete set of
FileMan directives for each virtual host, if any of the virtual
host configurations are different.
 
=head2 File Permissions

When using FileMan as a file manager, understanding and
implementing the file permissions is the hardest concept. First,
you need to have a good understanding of your operating system's
(OS) file permissions.

FileMan can allow groups of users to share the same web server
file space (tree), such that individuals can be prevented from
changing each others files and directories. An "admin" group can
also be specified, which allowes certain users to be able to
modify all the files and directories within the tree, as well
as, assign GID access to the files and directories.
 
File permissions are controlled by a group ID (GID) provided by
an authorization module for the user. It is assigned to the
files and directories that that user creates. 

An Apache environment variable must be set prior to each FileMan
request. This environment variable would normally be set by an
authorization module.

For example, the Apache::AuthzDBI module (presented above) can
provide an environment variable "REMOTE_GROUP" which contains
the group ID of the authorized user. The following FileMan
directive tells it which environment variable contains the
user's GID for the request:

    FileManOptions GIDEnv=REMOTE_GROUP

For example, if the authorization module sets the environment
variable:

	REMOTE_GROUP=1000

FileMan would set the GID for that user to 1000. If the GID is
valid (for Apache and it's OS), all files and directories created by
that user will have their GID set to 1000.

HINT:  If you set the "FileManOptions Debug 1" directive, the
environment variables will be listed along with other debuging
information.  You can then spot your GID environement variable
set by your authorization module in order to verify it's
existance and FileMan operation. 

An admin directive can also be specified which enables a user
with the specified admin GID to access and control all files and
directories within the current file manager directory (.XFM)
tree.

In summary, if the following directives are provided:

  FileManOptions GIDEnv=REMOTE_GROUP
  FileManOptions Admin=1000
 
The GIDEnv directive tells FileMan which environment variable
contains the GID (REMOTE_GROUP in this example). [This variable
would have been set by an authorization module.] If the GID for
the user happens to be 1000, then that user will have "admin"
privleges and it's commands (SetGID).

The operating system (OS) rules still apply to all of the GID
operations. For example (OS=UNIX), if Apache's program ID (PID)
is 100 and a file is owned by user 200, Apache can not change
the GID of file unless the Apache process is also a member of
the GID 200 group.

If a "group name" (instead of a number) is provided, the GID
name is looked-up in the /etc/group file in order to obtain the
numeric GID. This is very UNIX like and my not work for other
operating systems.

HINT: Any environment variable can be used to contain the
GID. Therefore, you can trick the authorization module into
coughing up a GID by using the REMOTE_USER (user) environment
variable and then simply create a group with the same name. 
Don't forget to make the Apache's process user ID (PUID) a
member of the group (in /etc/group). 

=head2 AutoIndex Functionality

When a .XFM directory is not present in the URL, FileMan will
function like AutoIndex. Note that the .XFM directory name can
be changed with a directive. This is explain later on in the
text.

=head1 DIRECTIVES

The display options (directives) are a composite of autoindex,
AutoIndex, and FileMan's own module directives.

The original module directives are maintained by FileMan, so
that any existing directives that you may have, can be used to
maintain the status quo.

=head2 autoindex DIRECTIVES

Apache normally comes with mod_autoindex C module. A number of
it's httpd.conf directives are provided when Apache is
installed.

Documentation for autoindex can be found at:

    http://www.apache.org/docs/mod/mod_autoindex.html

An incomplete (no Alt directives) and a very brief description
of the autoindex (used by Apache::Icon) directives is
provided below.

These directives are processed by Apache::Icon.pm which
provides icons to Apache::AutoIndex and Apache::FileMan.

=over

=item * FancyIndexing boolean

    The FancyIndexing directive tells FileMan to present a
    robust display which can include permissions, an icon, name,
    date, size, and description for each file and directory. All
    of the following autoindex and AutoIndex directives require
    FancyIndexing.
    
=item * HeaderName file file ...

    Inserts a list of files displayed at the top of the HTML
    page. After Apache 1.3.5 the filename can be a relative URI.

=item * IndexIgnore file file

    A list of files not to be displayed. The files can specify
    extensions, partial names, wild card expressions, or full
    filenames.  Multiple IndexIgnore directives add to the list.

=item * IndexOptions [+|-]option [+|-]option ... 
    
    There are several options. Please refer to the above URL:
	http://www.apache.org/docs/mod/mod_autoindex.html 
    for the complete list.

=item * IndexOrderDefault Ascending|Descending Name|Date|Size|Description

    IndexOrderDefault takes two arguments. The first must be
    either Ascending or Descending, indicating the direction of
    the sort. The second argument must be one of the keywords:
    Name, Date, Size, or Description. It identifies the primary
    sort key.

=item * ReadmeName file file ...

    A list of text files that will be displayed to the end of the
    HTML page.
    
=item * AddDescription "string" file file...

    The file description displayed for the given file (file name
    wild cards).

=item * AddIconByEncoding (alttext, url) MIME-encoding MIME-encoding ...
    
    The file icon (alttext, url) to be displayed according to
    the MIME-encoding (mime-encoding).
    
=item * AddIconByType (alttext, url) MIME-type MIME-type ...

    The file icon (alttext, url) to be displayed according to
    the MIME-type (mime-type).

=item * AddIcon (alttext, url) name name ...

    The file icon (alttext, url) to be displayed according to
    file name extension.

=item * DefaultIcon icon

    The file icon to be displayed if no other icon can be found.
    (default icon)

=back

=head2 AutoIndex DIRECTIVES

=over

=item * IndexOptions Thumbnails

    The listing will include thumbnails for pictures. Defaults to
    false.

=item * IndexOptions ShowPermissions
    
    Print file permissions. Defaults to false.

=item * PerlSetVar IndexHtmlBody 'expression'

    This is an expression that should produce complete <BODY>
    tag when eval'ed. An example:

=item * PerlSetVar IndexHtmlBody '<BODY BACKGROUND=\"$ENV{BACKGROUND}\">'

=item * PerlSetVar IndexHtmlTable value

    This is a string that will be inserted inside the table tag
    of the listing. For example: <TABLE $value>

=item * PerlSetVar IndexHtmlHead value

    This should be the url (absolute or relative) of a resource
    that would be inserted right after the <BODY> tag and just
    before anything else.

=item * PerlSetVar IndexHtmlFoot value

    This should be the url (absolute or relative) of a resource
    that would be inserted right before the ending </BODY> tag
    and after everything else.

=item * PerlSetVar IndexDebug [0|1]

    If set, the listing displayed will print debugging
    information. The default is 0.

=back

=head2 FileMan DIRECTIVES

=over

=item * FileManOptions Debug [0|1]

    If set to 1, the listing displayed will print debugging
    information. The default is 0.

=item * FileManOptions RootMenu command1 command2 . . .
    
    Sets the command menu for the Root directory. The default
    menu is: "Browse", "Upload", "Unzip", "Delete", "MkDir",
    "MkFile","Copy","Move","Edit","Rename","Help". If the first 
    command is preceded by '+' the following commands will be 
    added to the existing list or if preceded by '-' they well 
    be removed from the list. 

=item * FileManOptions HoldMenu command1 command2 . . .

    Sets the command menu for the Hold directory. The default
    menu is: "Browse", "Upload", "Unzip", "Delete", "MkDir",
    "MkFile","Copy","Move","Edit","Rename","Help". If the first 
    command is preceded by '+' the following commands will be 
    added to the existing list or if preceded by '-' they well 
    be removed from the list. 

=item * FileManOptions AdmnMenu command1 command2 . . .

    Sets the command menu for the admin routines. The default
    menu is: "SetGID", "Revoke", and "Debug".  If the first 
    command is preceded by '+' the following commands will be 
    added to the existing list or if preceded by '-' they well 
    be removed from the list. 

=item * FileManOptions Root Directory
    
    When operating in the AutoIndex mode, this option allows 
    you to specify the root directory where FileMan will not
    display the "Parent directory" item (the root).  The 
    string is compared with Perl regular expressions.

=item * FileManOptions RootDir Directory
    
    Sets the RootDir directory stub name from which the files
    can be managed. The default is ".XFM". You should probably
    consider changing this value to something else if you do not
    want people probing your web site. You may want to prefix
    the name with a '.' in order to hide it from view.
    
=item * FileManOptions HoldDir SubDirectory
    
    Set the hold subdirectory stub name of the where FileMan
    stores the Hold directory files. The default is "hold".
    
=item * FileManOptions TextLen n
    
    Sets the text entry field of the command form to length n.
    The default value is 49.  The "SetGID" text length is
    almost one-half this value (default 25).
     
=item * FileManOptions EditMax n
    
    Sets the maximum edit file byte size to n.  This is the
    maximum file size that can be edited.  The default value
    is 131072 bytes.
    
=item * FileManOptions PostMax n
    
    Sets the http maximum post byte size to n.  This is the
    maximum file size that can be uploaded.  The default value
    is 4,194,304 bytes.
    
=item * FileManOptions Admin n

    Sets the admin GID to n. If the user's GID equals the admin
    GID, the "SetGID" command will be provided and file access
    control will be provided for all files and directories in
    both the HOLD and ROOT directory trees.
    
=item * FileManOptions umask n
    
    Allows you to set the umask for the files and directories
    created.  Generally n is an octal number starting with a '0'.

=item * FileManOptions Help URL
    
    Sets the URL of the user help command.  The default URL is:
    http://www.xorgate.com/help/FileMan
    
=item * FileManOptions language [0|no|off]
    
    Tells FileMan not to use the Apache::Language module to
    translate messages. ('0', 'no', or 'off')  Defaults 'off'.
    When enabled the Apache::Language module must be loaded.
    Make sure if you set language on that you load the 
    negotiation module and either use the Multiviews option
    or the *.var method. 
    
=item * FileManOptions GIDEnv name    
        
    If an authorization module provides an environment variable
    (name) with the user's GID, the GIDEnv directive tells
    FileMan which variable contains the GID for the current
    request. The GID is then retrieved from the environment
    variable and is applied to the user's commands. For each
    command the source GID is checked to make sure that the GID
    matches each file and directory created. If a name (not a
    number) is provided, it is looked up in the /etc/group file
    to obtain the GID number.
    
=item * FileManOptions UserEnv name    
        
    An environment variable can be specified which holds the
    user name of the request.  If 'Basic' authorization is being
    used, the user name will be recovered from Apachei, regardless
    of what ever is specified for 'UserEnv name'.

=item * FileManOptions Revoke [1|0]
    
    A boolean value which tells FileMan to check the file
    "revoked" in the root fake directory (RootDir) for users and
    groups that will not be allowed to execute commands. This
    file is maintained by FileMan for the admin user through
    "Enable" and "Disable" commands provide in the Revoke form.
    Note that Apache will need to have read and write access in
    this file ("revoked") and root fake directory (.XFM).

=item * FileManOptions hold [1|0]
    
    A boolean value which tells FileMan to use and process the
    HOLD (hold) directory (tree), if it exists. ('1', 'yes', or
    'on') Default 0. If the HOLD directory does not exist, it
    will not use it :-).

=item * FileManOptions nocache [1|0]
    
    A boolean value which tells FileMan to have the expire time
    of the http header to zero so that browsers will not cache 
    FileMan's output. Default 0.

=back

=head1 THUMBNAILS

Generation of thumbnails is possible. This means that listing a
directory that contains images can be listed with little reduced
thumbnails beside each image name instead of the standard
'image' icon.

To enable this you simply need to preload Image::Macick in
Apache. The IndexOption option Thumbnails controls thumbnails
generation for specific directories like any other IndexOption
directive.

=head2 USAGE

The way thumbnails are generated/produced can be configured in
many ways.  A general overview of the procedure follows.

For each directory containing pictures, there will be a
.thumbnails directory created in it that will hold the thumbnails. Each
time the directory is accessed, and if thumbnail generation is
active, small thumbnails will be produced, shown beside each
image name, instead of the normal , generic, image icon.

That can be done in 2 ways. In the case the image is pretty
small, no actual thumbnail will be created. Instead the image
will resize the HEIGHT and WIDTH attributes of the IMG tag.

If the image is big enough, Image::Magick will resize it and
save (cache) it in the .thumbnails directory for the next
requests.

Changing configuration options will correctly refresh the cached
thumbnails. Also, if the original image is modified, the
thumbnail will be updated accordingly. Still, the browser might
screw things up if it preserves the cached images.  

The behavior of the Thumbnail generating code can be customized
with these PerlSetVar variables:

=head2 Thumbnail DIRECTIVES

=over

=item * IndexCacheDir dir

This is the name of the directory where the generated thumbnails
will be created.  Make sure the user under which the web server
runs has read and write permissions. Defaults to .thumbnails

=item * IndexCreateDir 0|1

Specifies that when a cache directory isn't found, should an
attempt to be made to create it. Defaults to 1(true), meaning if
possible, a missing cache directories will be created. 

=item * ThumbMaxFilesize bytes

This value fixes the maximum size of an image at which thumbnail
processing isn't even attempted.  Trying to process a few
very big images could bring a server down to it's knees.
Defaults to 500,000

=item * ThumbMinFilesize bytes

This value fixes the minumum size of an image at which thumbnail
processing isn't actually done. Since trying to process already
very small images could be an overkill, the image is simply
resized with the size attributes of the IMG tag. Defaults to
5,000.

=item * ThumbMaxWidth pixels

This value fixes the maximum x-size of an image at which
thumbnail processing isn't actually done. Since trying to
process already very small images would be an overkill, the
image is simply resized with the size attributes of the IMG tag.
Defaults to 4 times the default icon width.

=item * ThumbMaxHeight pixels

This value fixes the maximum y-size of an image at which
thumbnail processing isn't actually done. Since trying to
process already very small images would be an overkill, the
image is simply resized with the size attributes of the IMG tag.
Defaults to 4 times the default icon height

=item * ThumbScaleWidth scaling-factor

Preserved only if there is no scaling factor for the other axis
of the image. 

=item * ThumbScaleHeight scaling-factor

This value fixes an y-scaling factor between 0 and 1 to resize
the images. The image ratio will be preserved only if there is
no scaling factor for the other axis of the image. 

=item * ThumbWidth pixels

This value fixes a fixed x-dimension to resize the image. The
image ratio will be preserved only if there is no fixed scaling
factor for the other axis of the image. This has no effect if a
scaling factor is defined.

=item * ThumbHeight pixels

This value fixes a fixed x-dimension to resize the image. The
image ratio will be preserved only if there is no fixed scaling
factor for the other axis of the image. This has no effect if a
scaling factor is defined.

=back

=head1 TODO

Provide a means to allow external methods to be included in
the menus.

The thumbnail support needs to be tested. It was provide with
Apache:: AutoIndex, but I have not tested it yet.
    
Some minor changes to the thumbnails options will still have the
thumbnails regenerated. This should be avoided by checking the
attributes of the already existing thumbnail.

Some form of garbage collection should be performed on thumbnail
cache or the directories will fill up.

=head1 SEE ALSO

perl(1), L<Apache>(3), L<Apache::Icon>(3), L<Image::Magick>(3) .
L<Apache::AutoIndex>93)
    
=head1 SUPPORT

Please send any questions or comments to the Apache modperl 
mailing list <modperl@apache.org> or to me at <perler@xorgate.com>

=head1 NOTES

This code was made possible by :

=over

=item Philippe M. Chiasson

<gozer@ectoplasm.dyndns.com> Creator of Apache::AutoIndex.

=item Doug MacEachern 

<dougm@pobox.com>  Creator of Apache::Icon, and of course, mod_perl.

=item Rob McCool

Who produced the final mod_autoindex.c I copied, hrm.., well,
translated to perl.

=item The mod_perl mailing-list 

at <modperl@apache.org> for all your mod_perl related problems.

=back

=head1 AUTHOR

George Sanderson <george@xorgate.com>

=head1 COPYRIGHT

Copyright (c) 2000 George Sanderson All rights reserved. This
program is free software; you can redistribute it and/or modify
it under the same terms as Perl itself. 

Copyright (c) 1999 Philippe M. Chiasson. All rights reserved.
This program is free software; you can redistribute it and/or
modify it under the same terms as Perl itself. 

=cut
